from __future__ import annotations

from angr.sim_state import SimState
from .actions_mixin import ActionsMixinHigh, ActionsMixinLow
from .address_concretization_mixin import AddressConcretizationMixin
from .bvv_conversion_mixin import DataNormalizationMixin
from .clouseau_mixin import InspectMixinHigh
from .conditional_store_mixin import ConditionalMixin
from .convenient_mappings_mixin import ConvenientMappingsMixin
from .default_filler_mixin import DefaultFillerMixin, SpecialFillerMixin, ExplicitFillerMixin
from .dirty_addrs_mixin import DirtyAddrsMixin
from .hex_dumper_mixin import HexDumperMixin
from .label_merger_mixin import LabelMergerMixin
from .multi_value_merger_mixin import MultiValueMergerMixin
from .name_resolution_mixin import NameResolutionMixin
from .simplification_mixin import SimplificationMixin
from .simple_interface_mixin import SimpleInterfaceMixin
from .size_resolution_mixin import SizeNormalizationMixin, SizeConcretizationMixin
from .smart_find_mixin import SmartFindMixin
from .symbolic_merger_mixin import SymbolicMergerMixin
from .top_merger_mixin import TopMergerMixin
from .underconstrained_mixin import UnderconstrainedMixin
from .unwrapper_mixin import UnwrapperMixin

from .paged_memory.page_backer_mixins import ClemoryBackerMixin, ConcreteBackerMixin, DictBackerMixin
from .paged_memory.paged_memory_mixin import (
    PagedMemoryMixin,
    ListPagesMixin,
    UltraPagesMixin,
    ListPagesWithLabelsMixin,
    MVListPagesMixin,
    MVListPagesWithLabelsMixin,
)
from .paged_memory.privileged_mixin import PrivilegedPagingMixin
from .paged_memory.stack_allocation_mixin import StackAllocationMixin
from .paged_memory.paged_memory_multivalue_mixin import PagedMemoryMultiValueMixin
from .paged_memory.pages import (
    CooperationBase,
    MemoryObjectMixin,
    ISPOMixin,
    RefcountMixin,
    PermissionsMixin,
    HistoryTrackingMixin,
    PageBase,
    PageType,
    ListPage,
    MVListPage,
    UltraPage,
)

from .slotted_memory import SlottedMemoryMixin
from .regioned_memory import (
    RegionedMemoryMixin,
    RegionCategoryMixin,
    StaticFindMixin,
    AbstractMergerMixin,
    MemoryRegionMetaMixin,
    RegionedAddressConcretizationMixin,
)
from .keyvalue_memory_mixin import KeyValueMemoryMixin
from .javavm_memory_mixin import JavaVmMemoryMixin

# pylint:disable=missing-class-docstring


class DefaultMemory(
    HexDumperMixin,
    SmartFindMixin,
    UnwrapperMixin,
    NameResolutionMixin,
    DataNormalizationMixin,
    SimplificationMixin,
    InspectMixinHigh,
    ActionsMixinHigh,
    UnderconstrainedMixin,
    SizeConcretizationMixin,
    SizeNormalizationMixin,
    AddressConcretizationMixin,
    # InspectMixinLow,
    ActionsMixinLow,
    ConditionalMixin,
    ConvenientMappingsMixin,
    DirtyAddrsMixin,
    # -----
    StackAllocationMixin,
    ConcreteBackerMixin,
    ClemoryBackerMixin,
    DictBackerMixin,
    PrivilegedPagingMixin,
    UltraPagesMixin,
    DefaultFillerMixin,
    SymbolicMergerMixin,
    PagedMemoryMixin,
):
    pass


class DefaultListPagesMemory(
    HexDumperMixin,
    SmartFindMixin,
    UnwrapperMixin,
    NameResolutionMixin,
    DataNormalizationMixin,
    SimplificationMixin,
    ActionsMixinHigh,
    UnderconstrainedMixin,
    SizeConcretizationMixin,
    SizeNormalizationMixin,
    InspectMixinHigh,
    AddressConcretizationMixin,
    # InspectMixinLow,
    ActionsMixinLow,
    ConditionalMixin,
    ConvenientMappingsMixin,
    DirtyAddrsMixin,
    # -----
    StackAllocationMixin,
    ClemoryBackerMixin,
    DictBackerMixin,
    PrivilegedPagingMixin,
    ListPagesMixin,
    DefaultFillerMixin,
    SymbolicMergerMixin,
    PagedMemoryMixin,
):
    pass


class FastMemory(
    NameResolutionMixin,
    SimpleInterfaceMixin,
    SimplificationMixin,
    InspectMixinHigh,
    ConditionalMixin,
    ExplicitFillerMixin,
    DefaultFillerMixin,
    SlottedMemoryMixin,
):
    pass


class AbstractMemory(
    UnwrapperMixin,
    NameResolutionMixin,
    DataNormalizationMixin,
    SimplificationMixin,
    InspectMixinHigh,
    ActionsMixinHigh,
    UnderconstrainedMixin,
    SizeConcretizationMixin,
    SizeNormalizationMixin,
    # InspectMixinLow,
    ActionsMixinLow,
    ConditionalMixin,
    RegionedAddressConcretizationMixin,
    # -----
    RegionedMemoryMixin,
):
    pass


class RegionedMemory(
    RegionCategoryMixin,
    MemoryRegionMetaMixin,
    StaticFindMixin,
    UnwrapperMixin,
    NameResolutionMixin,
    DataNormalizationMixin,
    SimplificationMixin,
    SizeConcretizationMixin,
    SizeNormalizationMixin,
    AddressConcretizationMixin,
    ConvenientMappingsMixin,
    DirtyAddrsMixin,
    # -----
    ClemoryBackerMixin,
    DictBackerMixin,
    UltraPagesMixin,
    DefaultFillerMixin,
    AbstractMergerMixin,
    PagedMemoryMixin,
):
    pass


class LabeledMemory(
    SizeNormalizationMixin,
    ListPagesWithLabelsMixin,
    DefaultFillerMixin,
    TopMergerMixin,
    LabelMergerMixin,
    PagedMemoryMixin,
):
    """
    LabeledMemory is used in static analysis. It allows storing values with labels, such as `Definition`.
    """

    def _default_value(self, addr, size, **kwargs):  # pylint:disable=arguments-differ
        # TODO: Make _default_value() a separate Mixin

        if kwargs.get("name", "").startswith("merge_uc_"):
            # this is a hack. when this condition is satisfied, _default_value() is called inside Listpage.merge() to
            # create temporary values. we simply return a TOP value here.
            return self.state.top(size * self.state.arch.byte_width)

        # we never fill default values for non-existent loads
        kwargs["fill_missing"] = False
        return super()._default_value(addr, size, **kwargs)


class MultiValuedMemory(
    SizeNormalizationMixin,
    MVListPagesMixin,
    DefaultFillerMixin,
    MultiValueMergerMixin,
    PagedMemoryMixin,
    PagedMemoryMultiValueMixin,
):
    def _default_value(self, addr, size, **kwargs):  # pylint:disable=arguments-differ
        # TODO: Make _default_value() a separate Mixin

        if kwargs.get("name", "").startswith("merge_uc_"):
            # this is a hack. when this condition is satisfied, _default_value() is called inside Listpage.merge() to
            # create temporary values. we simply return a TOP value here.
            return self.state.top(size * self.state.arch.byte_width)

        # we never fill default values for non-existent loads
        kwargs["fill_missing"] = False
        return super()._default_value(addr, size, **kwargs)


class KeyValueMemory(
    KeyValueMemoryMixin,
):
    pass


class JavaVmMemory(
    JavaVmMemoryMixin,
):
    pass


SimState.register_default("sym_memory", DefaultMemory)
SimState.register_default("fast_memory", FastMemory)
SimState.register_default("abs_memory", AbstractMemory)
SimState.register_default("keyvalue_memory", KeyValueMemory)
SimState.register_default("javavm_memory", JavaVmMemory)


__all__ = (
    "AbstractMemory",
    "AbstractMergerMixin",
    "ActionsMixinHigh",
    "ActionsMixinLow",
    "AddressConcretizationMixin",
    "ClemoryBackerMixin",
    "ConcreteBackerMixin",
    "ConditionalMixin",
    "ConvenientMappingsMixin",
    "CooperationBase",
    "DataNormalizationMixin",
    "DefaultFillerMixin",
    "DefaultListPagesMemory",
    "DefaultMemory",
    "DictBackerMixin",
    "DirtyAddrsMixin",
    "ExplicitFillerMixin",
    "FastMemory",
    "HexDumperMixin",
    "HistoryTrackingMixin",
    "ISPOMixin",
    "InspectMixinHigh",
    "JavaVmMemory",
    "JavaVmMemoryMixin",
    "KeyValueMemory",
    "KeyValueMemoryMixin",
    "LabelMergerMixin",
    "LabeledMemory",
    "ListPage",
    "ListPagesMixin",
    "ListPagesWithLabelsMixin",
    "MVListPage",
    "MVListPagesMixin",
    "MVListPagesWithLabelsMixin",
    "MemoryObjectMixin",
    "MemoryRegionMetaMixin",
    "MultiValueMergerMixin",
    "MultiValuedMemory",
    "NameResolutionMixin",
    "PageBase",
    "PageType",
    "PagedMemoryMixin",
    "PagedMemoryMultiValueMixin",
    "PermissionsMixin",
    "PrivilegedPagingMixin",
    "RefcountMixin",
    "RegionCategoryMixin",
    "RegionedAddressConcretizationMixin",
    "RegionedMemory",
    "RegionedMemoryMixin",
    "SimpleInterfaceMixin",
    "SimplificationMixin",
    "SizeConcretizationMixin",
    "SizeNormalizationMixin",
    "SlottedMemoryMixin",
    "SmartFindMixin",
    "SpecialFillerMixin",
    "StackAllocationMixin",
    "StaticFindMixin",
    "SymbolicMergerMixin",
    "TopMergerMixin",
    "UltraPage",
    "UltraPagesMixin",
    "UnderconstrainedMixin",
    "UnwrapperMixin",
)



from __future__ import annotations
from typing import TYPE_CHECKING, TypeAlias
from collections import defaultdict
from collections.abc import MutableMapping

from angr.ailment.expression import StackBaseOffset
from angr.code_location import AILCodeLocation

if TYPE_CHECKING:
    from angr.analyses.decompiler.ssailification.ssailification import Def

# (stack offset | None, const value or offset from orig stack offset)
Value: TypeAlias = "set[tuple[int | None, int]]"


class TraversalState:
    """
    The abstract state for the traversal engine.
    """

    def __init__(
        self,
        arch,
        func,
        live_registers: MutableMapping[int, Value] | None = None,
        live_stackvars: MutableMapping[int, Value] | None = None,
        register_blackout: set[int] | None = None,
        live_vvars: MutableMapping[int, Value] | None = None,
        stackvar_bases: MutableMapping[int, tuple[int, int]] | None = None,
        register_bases: MutableMapping[int, tuple[int, int]] | None = None,
        stackvar_defs: MutableMapping[int, set[Def]] | None = None,
        register_defs: MutableMapping[int, set[Def]] | None = None,
        pending_ptr_defines_nonlocal_live: set[int] | None = None,
    ):
        self.arch = arch
        self.func = func

        self.register_blackout = set(register_blackout or ())
        self.live_registers = defaultdict(set, {} if live_registers is None else live_registers)
        self.live_stackvars = defaultdict(set, {} if live_stackvars is None else live_stackvars)
        self.live_vvars = defaultdict(set, {} if live_vvars is None else live_vvars)
        self.live_tmps: MutableMapping[int, Value] = defaultdict(
            set
        )  # tmps are internal to a block only and never propagated from another state

        self.stackvar_bases: MutableMapping[int, tuple[int, int]] = stackvar_bases if stackvar_bases is not None else {}
        self.register_bases: MutableMapping[int, tuple[int, int]] = register_bases if register_bases is not None else {}
        self.pending_ptr_defines: dict[int, list[tuple[AILCodeLocation, StackBaseOffset]]] = {}
        self.pending_ptr_defines_nonlocal_live = pending_ptr_defines_nonlocal_live or set()
        self.stackvar_defs = defaultdict(set, {} if stackvar_defs is None else stackvar_defs)
        self.register_defs = defaultdict(set, {} if register_defs is None else register_defs)

    def stackvar_unify(self, offset: int, size: int) -> tuple[int, int, set[int]]:
        seen = (offset, offset + size)
        queue = [(offset, offset + size)]
        popped: set[int] = set()
        while queue:
            offset, eoffset = queue.pop()
            for suboffset in range(offset, eoffset):
                noffset, nsize = self.stackvar_bases.get(suboffset, (suboffset, 0))
                neoffset = noffset + nsize
                if noffset < seen[0]:
                    queue.append((noffset, seen[0]))
                    seen = (noffset, seen[1])
                if neoffset > seen[1]:
                    queue.append((seen[1], neoffset))
                    seen = (seen[0], neoffset)
                if nsize != 0:
                    popped.add(noffset)

        final_offset, final_size = (seen[0], seen[1] - seen[0])
        for suboffset in range(*seen):
            self.stackvar_bases[suboffset] = (final_offset, final_size)

        return (final_offset, final_size, popped)

    def register_unify(self, offset: int, size: int) -> tuple[int, int, set[int]]:
        seen = (offset, offset + size)
        queue = [(offset, offset + size)]
        popped: set[int] = set()
        while queue:
            offset, eoffset = queue.pop()
            for suboffset in range(offset, eoffset):
                noffset, nsize = self.register_bases.get(suboffset, (suboffset, 0))
                neoffset = noffset + nsize
                if noffset < seen[0]:
                    queue.append((noffset, seen[0]))
                    seen = (noffset, seen[1])
                if neoffset > seen[1]:
                    queue.append((seen[1], neoffset))
                    seen = (seen[0], neoffset)
                if nsize != 0:
                    popped.add(noffset)

        final_offset, final_size = (seen[0], seen[1] - seen[0])
        for suboffset in range(*seen):
            self.register_bases[suboffset] = (final_offset, final_size)

        return (final_offset, final_size, popped)

    def copy(self) -> TraversalState:
        return TraversalState(
            self.arch,
            self.func,
            # these get copied
            live_registers=self.live_registers,
            live_stackvars=self.live_stackvars,
            register_blackout=self.register_blackout,
            live_vvars=self.live_vvars,
            pending_ptr_defines_nonlocal_live=set(self.pending_ptr_defines_nonlocal_live),
            stackvar_bases=dict(self.stackvar_bases),
            stackvar_defs={k: set(v) for k, v in self.stackvar_defs.items()},
            register_bases=dict(self.register_bases),
            register_defs={k: set(v) for k, v in self.register_defs.items()},
        )

    def merge(self, *others: TraversalState) -> bool:
        merge_occurred = False

        for o in others:
            for k, v in o.live_registers.items():
                merge_occurred |= bool(v.difference(self.live_registers[k]))
                self.live_registers[k].update(v)
            for k, v in o.live_stackvars.items():
                merge_occurred |= bool(v.difference(self.live_stackvars[k]))
                self.live_stackvars[k].update(v)
            for k, v in o.live_vvars.items():
                merge_occurred |= bool(v.difference(self.live_vvars[k]))
                self.live_vvars[k].update(v)
            for k0, (k1, s1) in o.stackvar_bases.items():
                k2, s2 = self.stackvar_bases.get(k0, (k0, 0))
                k3 = min(k1, k2)
                s3 = max(k1 + s1, k2 + s2) - k3
                if (k2, s2) != (k3, s3):
                    merge_occurred = True
                    self.stackvar_bases[k0] = (k3, s3)
            for k0, (k1, s1) in o.register_bases.items():
                k2, s2 = self.register_bases.get(k0, (k0, 0))
                k3 = min(k1, k2)
                s3 = max(k1 + s1, k2 + s2) - k3
                if (k2, s2) != (k3, s3):
                    merge_occurred = True
                    self.register_bases[k0] = (k3, s3)
            for k, d in o.stackvar_defs.items():
                merge_occurred |= bool(d.difference(self.stackvar_defs[k]))
                self.stackvar_defs[k].update(d)
            for k, d in o.register_defs.items():
                merge_occurred |= bool(d.difference(self.register_defs[k]))
                self.register_defs[k].update(d)
            merge_occurred |= bool(
                o.pending_ptr_defines_nonlocal_live.difference(self.pending_ptr_defines_nonlocal_live)
            )
            self.pending_ptr_defines_nonlocal_live.update(o.pending_ptr_defines_nonlocal_live)
            merge_occurred |= bool(o.register_blackout.difference(self.register_blackout))
            self.register_blackout.update(o.register_blackout)

        return merge_occurred



from __future__ import annotations

from .data import ArithmeticExpression, SpOffset, RegisterOffset
from .engine import (
    SimEngineLight,
    SimEngineLightVEX,
    SimEngineLightAIL,
    SimEngineNostmtVEX,
    SimEngineNostmtAIL,
    SimEngineNoexprAIL,
)

__all__ = (
    "ArithmeticExpression",
    "RegisterOffset",
    "SimEngineLight",
    "SimEngineLightAIL",
    "SimEngineLightVEX",
    "SimEngineNoexprAIL",
    "SimEngineNostmtAIL",
    "SimEngineNostmtVEX",
    "SpOffset",
)



# pylint:disable=missing-class-docstring,missing-function-docstring
"""
Base class for semantic variable naming patterns.

This module provides two base classes for semantic naming:
- ClinicNamingBase: For passes running in Clinic on AIL graphs
- RegionNamingBase: For passes running in RegionSimplifier on structured regions
"""

from __future__ import annotations
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
import logging

from angr.ailment.expression import Call, Const
from angr.sim_variable import SimVariable

if TYPE_CHECKING:
    import networkx

    from angr.ailment import Block
    from angr.knowledge_plugins.functions.function_manager import FunctionManager
    from angr.knowledge_plugins.variables.variable_manager import VariableManagerInternal
    from angr.analyses.decompiler.structuring.structurer_nodes import BaseNode

l = logging.getLogger(name=__name__)


class SemanticNamingBase(ABC):
    """
    Abstract base class for semantic variable naming patterns.

    Subclasses implement specific naming patterns (loop counters, array indices, etc.)
    This is the common base for both Clinic-based and RegionSimplifier-based passes.
    """

    # Priority determines order of application (lower = higher priority)
    # This matters when the same variable matches multiple patterns
    PRIORITY: int = 100

    def __init__(
        self,
        variable_manager: VariableManagerInternal,
        functions: FunctionManager,
    ):
        self._variable_manager = variable_manager
        self._functions = functions
        self._var_to_new_name: dict[SimVariable, str] = {}

    @abstractmethod
    def analyze(self) -> dict[SimVariable, str]:
        """
        Analyze and return a mapping of variables to their suggested names.

        :return: Dictionary mapping SimVariable to suggested name
        """
        raise NotImplementedError

    def apply_names(self, exclude_vars: set[SimVariable] | None = None) -> set[SimVariable]:
        """
        Apply the computed names to the variables.

        :param exclude_vars: Variables to skip (already named by higher priority patterns)
        :return: Set of variables that were renamed
        """
        exclude_vars = exclude_vars or set()
        renamed_vars: set[SimVariable] = set()

        for var, new_name in self._var_to_new_name.items():
            # Skip if already named by another pattern
            if var in exclude_vars:
                continue

            # Check if variable already has a meaningful name (not auto-generated)
            if var.renamed:
                continue

            # Find the unified variable and rename it
            unified_var = self._variable_manager.unified_variable(var)
            target_var = unified_var if unified_var is not None else var

            # Also skip if unified var is in exclude set
            if unified_var in exclude_vars:
                continue

            if unified_var in renamed_vars:
                continue

            if target_var.renamed:
                continue

            l.debug("Renaming %s -> %s (pattern: %s)", target_var.name, new_name, self.__class__.__name__)
            target_var.name = new_name
            target_var.renamed = True
            target_var.clear_hash()
            renamed_vars.add(var)
            if unified_var:
                renamed_vars.add(unified_var)

        return renamed_vars

    # --- Helper methods for subclasses ---

    def _get_linked_variable(self, expr) -> SimVariable | None:
        """
        Get the SimVariable linked to an expression, if any.
        """
        if hasattr(expr, "variable") and expr.variable is not None:
            return self._variable_manager.unified_variable(expr.variable)
        return None

    def _get_function_name(self, call: Call) -> str | None:
        """
        Extract the function name from a Call expression.
        """
        target = call.target

        # Direct function address
        if isinstance(target, Const) and self._functions.contains_addr(target.value_int):
            func = self._functions.get_by_addr(target.value_int)
            return func.name
        if isinstance(target, str):
            return target
        return None

    @staticmethod
    def _normalize_name(name: str) -> str:
        """Normalize a function name for matching."""
        return name.lower().strip("_")


class ClinicNamingBase(SemanticNamingBase):
    """
    Base class for semantic naming passes that run in Clinic on AIL graphs.

    These passes operate on the raw AIL graph before structuring, allowing them
    to analyze control flow patterns directly.
    """

    def __init__(
        self,
        ail_graph: networkx.DiGraph,
        variable_manager: VariableManagerInternal,
        functions: FunctionManager,
        entry_node: Block,
    ):
        super().__init__(variable_manager, functions)
        self._graph = ail_graph
        self._entry_node = entry_node


class RegionNamingBase(SemanticNamingBase):
    """
    Base class for semantic naming passes that run in RegionSimplifier on structured regions.

    These passes operate after structuring, allowing them to leverage structured information
    like LoopNode, ConditionNode, etc. They can reuse loop analysis results from structuring
    instead of re-analyzing the graph.
    """

    def __init__(
        self,
        region: BaseNode,
        variable_manager: VariableManagerInternal,
        functions: FunctionManager,
    ):
        super().__init__(variable_manager, functions)
        self._region = region



# pylint: disable=missing-class-docstring
from __future__ import annotations
import glob
import importlib
import importlib.resources
import os
import shutil
import subprocess
import sys
from distutils.command.build import build as st_build

from setuptools import Command, setup
from setuptools.command.develop import develop as st_develop
from setuptools.errors import LibError

# Import setuptools_rust to ensure an error is raised if not installed
try:
    _ = importlib.import_module("setuptools_rust")
except ImportError as err:
    raise Exception("angr requires setuptools-rust to build") from err

if sys.platform == "darwin":
    library_file = "unicornlib.dylib"
elif sys.platform in ("win32", "cygwin"):
    library_file = "unicornlib.dll"
else:
    library_file = "unicornlib.so"


def build_unicornlib():
    try:
        importlib.import_module("pyvex")
    except ImportError as e:
        raise LibError("You must install pyvex before building angr") from e

    env = os.environ.copy()
    env_data = (
        ("PYVEX_INCLUDE_PATH", "pyvex", "include"),
        ("PYVEX_LIB_PATH", "pyvex", "lib"),
        ("PYVEX_LIB_FILE", "pyvex", "lib\\pyvex.lib"),
    )
    for var, pkg, fnm in env_data:
        base = importlib.resources.files(pkg)
        for child in fnm.split("\\"):
            base = base.joinpath(child)
        env[var] = str(base)

    if sys.platform == "win32":
        cmd = ["nmake", "/f", "Makefile-win"]
    elif shutil.which("gmake") is not None:
        cmd = ["gmake"]
    else:
        cmd = ["make"]
    try:
        subprocess.run(cmd, cwd="native/unicornlib", env=env, check=True)
    except FileNotFoundError as err:
        raise LibError("Couldn't find " + cmd[0] + " in PATH") from err
    except subprocess.CalledProcessError as err:
        raise LibError("Error while building unicornlib: " + str(err)) from err

    shutil.rmtree("angr/lib", ignore_errors=True)
    os.mkdir("angr/lib")
    shutil.copy(os.path.join("native/unicornlib", library_file), "angr")


def clean_unicornlib():
    oglob = glob.glob("native/*.o")
    oglob += glob.glob("native/*.obj")
    oglob += glob.glob("native/*.so")
    oglob += glob.glob("native/*.dll")
    oglob += glob.glob("native/*.dylib")
    for fname in oglob:
        os.unlink(fname)


class build(st_build):
    def run(self, *args):
        self.execute(build_unicornlib, (), msg="Building unicornlib")
        super().run(*args)


class clean(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.execute(clean, (), msg="Cleaning unicornlib")


class develop(st_develop):
    def run(self):
        self.run_command("build")
        super().run()


cmdclass = {
    "build": build,
    "clean_unicornlib": clean,
    "develop": develop,
}


try:
    from setuptools.command.editable_wheel import editable_wheel as st_editable_wheel

    class editable_wheel(st_editable_wheel):
        def run(self):
            self.run_command("build")
            super().run()

    cmdclass["editable_wheel"] = editable_wheel
except ModuleNotFoundError:
    pass


setup(cmdclass=cmdclass)



from __future__ import annotations

from . import ux
from . import autoimport
from .loggers import Loggers
from .plugins import PluginHub, PluginPreset
from .hookset import HookSet
from .picklable_lock import PicklableLock

__all__ = (
    "HookSet",
    "Loggers",
    "PicklableLock",
    "PluginHub",
    "PluginPreset",
    "autoimport",
    "ux",
)



from __future__ import annotations

import contextlib
import logging
from collections import defaultdict
from collections.abc import Sequence
from typing import Any

import archinfo
import pypcode
import pyvex

from . import Analysis
from angr.analyses import AnalysesHub
from angr.engines import pcode
from angr.errors import AngrTypeError
from angr.knowledge_plugins import Function
from angr.utils.library import get_cpp_function_name
from angr.utils.formatting import ansi_color_enabled, ansi_color, add_edge_to_buffer
from angr.block import DisassemblerInsn, CapstoneInsn, SootBlockNode
from angr.codenode import BlockNode
from .disassembly_utils import decode_instruction

IRSBType = pyvex.IRSB | pcode.lifter.IRSB
IROpObjType = pyvex.stmt.IRStmt | pypcode.PcodeOp

l = logging.getLogger(name=__name__)

# pylint: disable=unidiomatic-typecheck


class DisassemblyPiece:
    addr = None
    ident = float("nan")

    def render(self, formatting=None) -> list[str]:
        x = self._render(formatting)
        if len(x) == 1:
            return [self.highlight(x[0], formatting)]
        return x

    def _render(self, formatting):
        raise NotImplementedError

    def getpiece(self, formatting, column):  # pylint:disable=unused-argument
        return self

    def width(self, formatting):
        r = self._render(formatting)
        if not r:
            return 0
        return max(len(x) for x in r)

    def height(self, formatting):
        return len(self._render(formatting))

    @staticmethod
    def color(string, coloring, formatting):
        try:
            return "{}{}{}".format(formatting["colors"][coloring][0], string, formatting["colors"][coloring][1])
        except KeyError:
            return string

    def highlight(self, string, formatting=None):
        try:
            if formatting is not None:
                if "format_callback" in formatting:
                    return formatting["format_callback"](self, string)
                if self in formatting["highlight"]:
                    return self.color(string, "highlight", formatting)
        except KeyError:
            pass
        return string

    def __eq__(self, other) -> bool:
        return False


class FunctionStart(DisassemblyPiece):
    def __init__(self, func):
        """
        Constructor.

        :param angr.knowledge.Function func: The function instance.
        """

        self.addr = func.addr
        self.vars = []
        self.name = func.name
        self.is_simprocedure = func.is_simprocedure
        self.sim_procedure = None
        if func.is_syscall:
            self.sim_procedure = func._project.simos.syscall_from_addr(self.addr)
        elif func.is_simprocedure:
            self.sim_procedure = func._project.hooked_by(self.addr)

    def _render(self, formatting):
        # TODO: Make the individual elements be individual Pieces
        return [f"{name} = {offset:#x}" for offset, name in self.vars]

    def height(self, formatting):
        return len(self.vars)


class Label(DisassemblyPiece):
    def __init__(self, addr, name):
        self.addr = addr
        self.name = name

    def _render(self, formatting):  # pylint:disable=unused-argument
        return [self.name + ":"]


class IROp(DisassemblyPiece):
    __slots__ = (
        "addr",
        "irsb",
        "obj",
        "seq",
    )

    addr: int
    seq: int
    obj: IROpObjType
    irsb: IRSBType

    def __init__(self, addr: int, seq: int, obj: IROpObjType, irsb: IRSBType):
        self.addr = addr
        self.seq = seq
        self.obj = obj
        self.irsb = irsb

    def __str__(self):
        if pcode and isinstance(self.obj, pypcode.PcodeOp):
            return pypcode.PcodePrettyPrinter.fmt_op(self.obj)
        return str(self.obj)

    def _render(self, formatting):  # pylint:disable=unused-argument
        return [str(self)]


class BlockStart(DisassemblyPiece):
    def __init__(self, block, parentfunc, project):
        self.addr = block.addr
        self.size = block.size
        self.parentfunc = parentfunc
        self.project = project

    def _render(self, formatting):
        return []


class Hook(DisassemblyPiece):
    def __init__(self, block):
        self.addr = block.addr
        simproc_name = str(block.sim_procedure)
        self.name = simproc_name.split()[-1].strip("'<>")
        self.short_name = simproc_name.strip("'<>").split(".")[-1]

    def _render(self, formatting):
        return ["SimProcedure " + self.short_name]

    def __eq__(self, other):
        return type(other) is Hook and self.name == other.name


class Instruction(DisassemblyPiece):
    def __init__(self, insn, parentblock, project=None):
        self.addr = insn.address
        self.size = insn.size
        self.insn = insn
        self.parentblock = parentblock
        self.project = parentblock.project if parentblock is not None else project
        self.arch = self.project.arch
        self.format = ""
        self.components = ()
        self.opcode: Opcode = None  # type: ignore
        self.operands = []

        # the following members will be filled in after dissecting the instruction
        self.type = None
        self.branch_type = None
        self.branch_target_operand = None

        self.dissect_instruction()

        if isinstance(insn, CapstoneInsn):
            decode_instruction(self.arch, self)

    @property
    def mnemonic(self):
        return self.opcode

    def reload_format(self):
        self.insn = CapstoneInsn(next(self.arch.capstone.disasm(self.insn.bytes, self.addr)))
        self.dissect_instruction()

    def dissect_instruction(self):
        if isinstance(
            self.arch,
            (archinfo.ArchAArch64, archinfo.ArchARM, archinfo.ArchARMEL, archinfo.ArchARMHF, archinfo.ArchARMCortexM),
        ):
            self.dissect_instruction_for_arm()
        else:
            # the default one works well for x86, add more arch-specific
            # code when you find it doesn't meet your need.
            self.dissect_instruction_by_default()

    def dissect_instruction_for_arm(self):
        self.opcode = Opcode(self)
        self.operands = []

        # We use capstone for arm64 disassembly, so this assertion must success
        assert hasattr(self.insn, "operands")

        op_str = self.insn.op_str
        dummy_operands = self.split_arm_op_string(op_str)

        for operand in dummy_operands:
            opr_pieces = self.split_op_string(operand)
            cur_operand = []

            if not (operand and opr_pieces):
                # opr_pieces may contain empty string when invalid disasm
                # result is generated by capstone
                l.error(f'Failed to parse insn "{self.insn}". Please report.')
                self.operands.clear()
                break

            if opr_pieces[0][0].isalpha() and opr_pieces[0] in self.arch.registers:
                cur_operand.append(Register(opr_pieces[0]))
                # handle register's suffix (e.g. "sp!", "d0[1]", "v0.16b")
                cur_operand.extend(opr_pieces[1:])
                self.operands.append(cur_operand)
                continue

            for i, p in enumerate(opr_pieces):
                if p[0].isnumeric():
                    if any(
                        (
                            i > 0 and opr_pieces[i - 1] == ".",
                            i > 1
                            and (
                                opr_pieces[i - 2] in ["lsl", "lsr", "asr", "ror", "msl"]
                                or opr_pieces[i - 2][:3] in ("uxt", "sxt")
                            ),
                        )
                    ):
                        cur_operand.append(p)
                        continue
                    # Always set False. I don't see any '+' sign appear
                    # in capstone's arm disasm result
                    with_sign = False
                    try:
                        v = int(p, 0)
                    except ValueError:
                        l.error("Failed to parse operand %s at %016x. Please report.", p, self.addr)
                        cur_operand.append(p)
                        continue
                    if i > 0 and opr_pieces[i - 1] == "-":
                        v = -v
                        cur_operand.pop()
                    with_pound_sign = False
                    if i > 0 and opr_pieces[i - 1] == "#":
                        with_pound_sign = True
                        cur_operand.pop()
                    cur_operand.append(Value(v, with_sign, render_with_pound_sign=with_pound_sign))
                elif p[0].isalpha() and p in self.arch.registers:
                    cur_operand.append(Register(p))
                else:
                    cur_operand.append(p)
            self.operands.append(cur_operand)

        for i, opr in enumerate(self.operands):
            # set extra dummy operand type to default 0
            op_type = self.insn.operands[i].type if i < len(self.insn.operands) else 0
            self.operands[i] = Operand.build(op_type, i, opr, self)

        if len(self.operands) == 0 and len(self.insn.operands) != 0:
            l.error("Operand parsing failed for instruction %s at address %x", str(self.insn), self.insn.address)
            return

    @staticmethod
    def split_arm_op_string(op_str: str):
        # Split arm operand string with commas outside the square brackets
        pieces = []
        in_square_brackets = False
        cur_opr = ""
        for c in op_str:
            if c == "[":
                in_square_brackets = True
            if c == "]":
                in_square_brackets = False
            if c == "," and not in_square_brackets:
                pieces.append(cur_opr)
                cur_opr = ""
                continue
            if c == " ":
                continue
            cur_opr += c
        if cur_opr:
            pieces.append(cur_opr)
        return pieces

    def dissect_instruction_by_default(self):
        # perform a "smart split" of an operands string into smaller pieces
        insn_pieces = self.split_op_string(self.insn.op_str)
        self.operands = []
        cur_operand = None
        i = len(insn_pieces) - 1
        cs_op_num = -1
        nested_mem = False

        # iterate over operands in reverse order
        while i >= 0:
            c = insn_pieces[i]
            if c == "":
                i -= 1
                continue

            if cur_operand is None:
                cur_operand = []
                self.operands.append(cur_operand)

            # Check if this is a number or an identifier.
            ordc = ord(c[0])
            # pylint:disable=too-many-boolean-expressions
            if 0x30 <= ordc <= 0x39 or 0x41 <= ordc <= 0x5A or 0x61 <= ordc <= 0x7A:
                # perform some basic classification
                intc = None
                reg = False
                try:
                    intc = int(c, 0)
                except ValueError:
                    reg = c in self.arch.registers

                # if this is a "live" piece, liven it up!
                # special considerations:
                # - registers should consolidate with a $ or % prefix
                # - integers should consolidate with a sign prefix

                if reg:
                    prefix = ""
                    if i > 0 and insn_pieces[i - 1] in ("$", "%"):
                        prefix = insn_pieces[i - 1]
                        insn_pieces[i - 1] = ""
                    cur_operand.append(Register(c, prefix))
                elif intc is not None:
                    with_sign = False
                    if i > 0 and insn_pieces[i - 1] in ("+", "-"):
                        with_sign = True
                        if insn_pieces[i - 1] == "-":
                            intc = -intc  # pylint: disable=invalid-unary-operand-type
                        insn_pieces[i - 1] = ""
                    cur_operand.append(Value(intc, with_sign))
                else:
                    cur_operand.append(c)

            elif c == "," and not nested_mem:
                cs_op_num -= 1
                cur_operand = None

            elif c == ":":  # XXX this is a hack! fix this later
                insn_pieces[i - 1] += ":"

            else:
                # Check if we are inside braces or parentheses. Do not forget
                # that we are iterating in reverse order!
                if c == "]" or c == ")":
                    nested_mem = True

                elif c == "[" or c == "(":
                    nested_mem = False

                if cur_operand is None:
                    cur_operand = [c]
                    self.operands.append(cur_operand)
                else:
                    cur_operand.append(c if c[0] != "," else c + " ")

            i -= 1

        self.opcode = Opcode(self)
        self.operands.reverse()

        if not hasattr(self.insn, "operands"):
            # Not all disassemblers provide operands. Just use our smart split
            for i, o in enumerate(self.operands):
                o.reverse()
                self.operands[i] = Operand.build(1, i, o, self)
            return

        if len(self.operands) != len(self.insn.operands):
            l.error(
                "Operand parsing failed for instruction %s. %d operands are parsed, while %d are expected.",
                str(self.insn),
                len(self.operands),
                len(self.insn.operands),
            )
            self.operands = []
            return

        for i, o in enumerate(self.operands):
            o.reverse()
            self.operands[i] = Operand.build(self.insn.operands[i].type, i, o, self)

    @staticmethod
    def split_op_string(insn_str):
        pieces = []
        in_word = False
        for c in insn_str:
            if c.isspace():
                in_word = False
                continue
            if c.isalnum():
                if in_word:
                    pieces[-1] += c
                else:
                    in_word = True
                    pieces.append(c)
            else:
                in_word = False
                pieces.append(c)
        return pieces

    def _render(self, formatting=None):
        return [
            "{} {}".format(self.opcode.render(formatting)[0], ", ".join(o.render(formatting)[0] for o in self.operands))
        ]


class SootExpression(DisassemblyPiece):
    def __init__(self, expr):
        self.expr = expr

    def _render(self, formatting=None):
        return [self.expr]


class SootExpressionTarget(SootExpression):
    def __init__(self, target_stmt_idx):
        super().__init__(target_stmt_idx)
        self.target_stmt_idx = target_stmt_idx

    def _render(self, formatting=None):
        return [f"Goto {self.target_stmt_idx}"]


class SootExpressionStaticFieldRef(SootExpression):
    def __init__(self, field):
        field_str = ".".join(field)
        super().__init__(field_str)
        self.field = field
        self.field_str = field_str

    def _render(self, formatting=None):
        return [self.field_str]


class SootExpressionInvoke(SootExpression):
    Virtual = "virtual"
    Static = "static"
    Special = "special"

    def __init__(self, invoke_type, expr):
        super().__init__(str(expr))

        self.invoke_type = invoke_type
        self.base = str(expr.base) if self.invoke_type in (self.Virtual, self.Special) else ""
        self.method_name = expr.method_name
        self.arg_str = expr.list_to_arg_str(expr.args)

    def _render(self, formatting=None):
        return [
            "{}{}({}) [{}]".format(
                self.base + "." if self.base else "", self.method_name, self.arg_str, self.invoke_type
            )
        ]


class SootStatement(DisassemblyPiece):
    def __init__(self, block_addr, raw_stmt):
        self.addr = block_addr.copy()
        self.addr.stmt_idx = raw_stmt.label
        self.raw_stmt = raw_stmt

        self.components = []

        self._parse()

    @property
    def stmt_idx(self):
        return self.addr.stmt_idx

    def _parse(self):
        func = f"_parse_{self.raw_stmt.__class__.__name__}"

        if hasattr(self, func):
            getattr(self, func)()
        else:
            # print func
            self.components += [f"NotImplemented: {func}"]

    def _expr(self, expr):
        func = f"_handle_{expr.__class__.__name__}"

        if hasattr(self, func):
            return getattr(self, func)(expr)
        # print func
        return SootExpression(str(expr))

    def _render(self, formatting=None):
        return [
            " ".join(
                [
                    component if type(component) is str else component.render(formatting=formatting)[0]
                    for component in self.components
                ]
            )
        ]

    #
    # Statement parsers
    #

    def _parse_AssignStmt(self):
        self.components += [
            SootExpression(str(self.raw_stmt.left_op)),
            "=",
            self._expr(self.raw_stmt.right_op),
        ]

    def _parse_InvokeStmt(self):
        self.components += [
            self._expr(self.raw_stmt.invoke_expr),
        ]

    def _parse_GotoStmt(self):
        self.components += [
            SootExpressionTarget(self.raw_stmt.target),
        ]

    def _parse_IfStmt(self):
        self.components += [
            "if (",
            SootExpression(str(self.raw_stmt.condition)),
            ")",
            SootExpressionTarget(self.raw_stmt.target),
        ]

    def _parse_ReturnVoidStmt(self):
        self.components += [
            "return",
        ]

    def _parse_IdentityStmt(self):
        self.components += [
            SootExpression(str(self.raw_stmt.left_op)),
            "<-",
            SootExpression(str(self.raw_stmt.right_op)),
        ]

    #
    # Expression handlers
    #

    def _handle_SootStaticFieldRef(self, expr):
        return SootExpressionStaticFieldRef(expr.field[::-1])

    def _handle_SootVirtualInvokeExpr(self, expr):
        return SootExpressionInvoke(SootExpressionInvoke.Virtual, expr)

    def _handle_SootStaticInvokeExpr(self, expr):
        return SootExpressionInvoke(SootExpressionInvoke.Static, expr)

    def _handle_SootSpecialInvokeExpr(self, expr):
        return SootExpressionInvoke(SootExpressionInvoke.Special, expr)


class Opcode(DisassemblyPiece):
    def __init__(self, parentinsn):
        self.addr = parentinsn.addr
        self.insn = parentinsn.insn
        self.parentinsn = parentinsn
        self.opcode_string = self.insn.mnemonic
        self.ident = (self.addr, "opcode")

    def _render(self, formatting=None):
        return [self.opcode_string.ljust(7)]

    def __eq__(self, other):
        return type(other) is Opcode and self.opcode_string == other.opcode_string


class Operand(DisassemblyPiece):
    def __init__(self, op_num, children, parentinsn):
        self.addr = parentinsn.addr
        self.children: list = children
        self.parentinsn = parentinsn
        self.op_num = op_num
        self.ident = (self.addr, "operand", self.op_num)

        for i, c in enumerate(self.children):
            if type(c) not in (bytes, str):
                c.ident = (self.addr, "operand piece", self.op_num, i)
                c.parentop = self

    @property
    def cs_operand(self):
        return self.parentinsn.insn.operands[self.op_num]

    def _render(self, formatting):
        return [
            "".join(
                x if type(x) is str else x.decode() if type(x) is bytes else x.render(formatting)[0]
                for x in self.children
            )
        ]

    @staticmethod
    def build(operand_type, op_num, children, parentinsn):
        # Maps capstone operand types to operand classes
        MAPPING = {
            0: Operand,  # default type for operand that haven't been fully implemented
            1: RegisterOperand,
            2: ConstantOperand,
            3: MemoryOperand,
            4: Operand,  # ARM FP
            64: Operand,  # ARM CIMM
            65: Operand,  # ARM PIMM   | ARM64 REG_MRS
            66: Operand,  # ARM SETEND | ARM64 REG_MSR
            67: Operand,  # ARM SYSREG | ARM64 PSTATE
            68: Operand,  # ARM64 SYS
            69: Operand,  # ARM64 PREFETCH
            70: Operand,  # ARM64 BARRIER
        }

        cls = MAPPING.get(operand_type)
        if cls is None:
            raise ValueError(f"Unknown capstone operand type {operand_type}.")

        operand = cls(op_num, children, parentinsn)

        # Post-processing
        if cls is MemoryOperand:
            operand = Operand._post_process_memory_operand(parentinsn, operand)

        return operand

    @staticmethod
    def _post_process_memory_operand(parentinsn: Instruction, operand: MemoryOperand) -> Operand:
        archname = parentinsn.arch.name
        if archname in {"AMD64", "ARM", "ARMEL", "ARMHF", "ARMCortexM"} and len(operand.values) == 2:
            op0, op1 = operand.values
            if type(op0) is Register and op0.is_ip and type(op1) is Value:
                # Indirect addressing in x86-64 and ARM
                # 400520  push [rip+0x200782] ==>  400520  push [0x600ca8]
                # 80410e6  ldr r1, [pc, #0x98]  ==>   80410e6  ldr r1, [0x8041182]
                match archname:
                    case "AMD64":
                        absolute_addr = parentinsn.addr + parentinsn.size + op1.val
                    case "ARM" | "ARMEL" | "ARMHF" | "ARMCortexM":
                        if parentinsn.addr & 1:
                            # thumb mode
                            absolute_addr = (parentinsn.addr & 0xFFFF_FFFE) + 4 + op1.val
                        else:
                            # arm mode
                            absolute_addr = parentinsn.addr + 8 + op1.val
                    case _:
                        raise AngrTypeError(f"Unsupported architecture {archname} for post-processing memory operand.")
                return MemoryOperand(1, [*operand.prefix, "[", Value(absolute_addr, False), "]"], parentinsn)
        return operand


class ConstantOperand(Operand):
    pass


class RegisterOperand(Operand):
    @property
    def register(self):
        return next((child for child in self.children if isinstance(child, Register)), None)

    def _render(self, formatting):
        custom_value_str = None
        if formatting is not None:
            with contextlib.suppress(KeyError):
                custom_value_str = formatting["custom_values_str"][self.ident]

        if custom_value_str:
            return [custom_value_str]
        return super()._render(formatting)


class MemoryOperand(Operand):
    def __init__(self, op_num, children, parentinsn):
        super().__init__(op_num, children, parentinsn)

        # a typical "children" looks like the following:
        # [ 'dword', 'ptr', '[', Register, Value, ']' ]
        # or
        # [ '[', Register, ']' ]
        # or
        # [ Value, '(', Register, ')' ]
        # or
        # [ Register, ",", Value]

        # it will be converted into more meaningful and Pythonic properties

        self.segment_selector = None
        self.prefix = []
        self.suffix_str = ""  # could be arm pre index mark "!"
        self.values: list[str | DisassemblyPiece] = []
        self.offset = []
        # offset_location
        # - prefix: -0xff00($gp)
        # - before_value: 0xff00+rax
        # - after_value: rax+0xff00
        self.offset_location = "after_value"
        # values_style
        # - square: [rax+0x10]
        # - curly: {rax+0x10}
        # - paren: (rax+0x10)
        self.values_style = "square"
        # separator (between register and value)
        # - "": rax+0x10
        # - "comma": r0, #0x10
        self.separator = ""

        try:
            if "[" in self.children:
                self._parse_memop_squarebracket()
            elif "(" in self.children:
                self._parse_memop_paren()
            else:
                raise ValueError

        except ValueError:
            l.error("Failed to parse operand children %s. Please report to Fish.", self.children)

            # setup all dummy properties
            self.prefix = []
            self.values = []

    def _parse_memop_squarebracket(self):
        if self.children[0] != "[":
            try:
                square_bracket_pos = self.children.index("[")
            except ValueError:  # pylint: disable=try-except-raise
                raise

            self.prefix = self.children[:square_bracket_pos]

            # take out segment selector
            if len(self.prefix) == 3:
                self.segment_selector = self.prefix[-1]
                self.prefix = self.prefix[:-1]
            else:
                self.segment_selector = None

        else:
            # empty
            square_bracket_pos = 0
            self.prefix = []
            self.segment_selector = None

        close_square_pos = len(self.children) - 1
        if self.children[-1] != "]":
            if self.children[-1] == "!" and self.children[-2] == "]":
                # arm64 pre index
                self.suffix_str = "!"
                close_square_pos -= 1
            else:
                raise ValueError

        self.values = self.children[square_bracket_pos + 1 : close_square_pos]
        if len(self.values) >= 3 and self.values[1] == ",":
            # arm style memory operand: [r0, #0x10], or [pc, r3, lsl #1]
            self.separator = "comma"
            # remove all commas and consolidate all remaining strings
            new_values = []
            last_item = None
            for v in self.values:
                if v == ",":
                    if last_item is not None:
                        new_values.append(last_item)
                        last_item = None
                elif isinstance(v, str):
                    last_item = v if last_item is None else last_item + v
                else:
                    if last_item is not None:
                        new_values.append(last_item)
                        last_item = None
                    new_values.append(v)
            if last_item is not None:
                new_values.append(last_item)
            self.values = new_values

    def _parse_memop_paren(self):
        offset = []
        self.values_style = "paren"

        if self.children[0] != "(":
            try:
                paren_pos = self.children.index("(")
            except ValueError:  # pylint: disable=try-except-raise
                raise

            if all(isinstance(item, str) for item in self.children[:paren_pos]):
                # parse prefix
                self.prefix = self.children[:paren_pos]
            elif all(isinstance(item, Value) for item in self.children[:paren_pos]):
                # parse offset
                # force each piece to be rendered with its sign (+/-)
                offset += self.children[:paren_pos]
                # offset appears before the left parenthesis
                self.offset_location = "prefix"

        else:
            paren_pos = 0
            self.prefix = []
            self.segment_selector = None

        self.values = self.children[paren_pos + 1 : len(self.children) - 1]
        self.offset = offset

    def _render(self, formatting):
        if self.prefix is None:
            # we failed in parsing. use the default rendering
            return super()._render(formatting)
        values_style = self.values_style
        show_prefix = True
        custom_values_str = None

        if formatting is not None:
            with contextlib.suppress(KeyError):
                values_style = formatting["values_style"][self.ident]

            try:
                show_prefix_str = formatting["show_prefix"][self.ident]
                if show_prefix_str in ("false", "False"):
                    show_prefix = False
            except KeyError:
                pass

            with contextlib.suppress(KeyError):
                custom_values_str = formatting["custom_values_str"][self.ident]

        prefix_str = " ".join(self.prefix) + " " if show_prefix and self.prefix else ""
        if custom_values_str is not None:
            value_str = custom_values_str
        else:
            sep = "," if self.separator == "comma" else ""
            value_str = sep.join(x.render(formatting)[0] if not isinstance(x, str) else x for x in self.values)

        if values_style == "curly":
            left_paren, right_paren = "{", "}"
        elif values_style == "paren":
            left_paren, right_paren = "(", ")"
        else:  # square
            left_paren, right_paren = "[", "]"

        if self.offset:
            offset_str = "".join(x.render(formatting)[0] if not isinstance(x, str) else x for x in self.offset)

            # combine values and offsets according to self.offset_location
            if self.offset_location == "prefix":
                value_str = f"{offset_str}{left_paren}{value_str}{right_paren}"
            elif self.offset_location == "before_value":
                value_str = f"{left_paren}{offset_str}{value_str}{right_paren}"
            else:  # after_value
                value_str = f"{left_paren}{value_str}{offset_str}{right_paren}"
        else:
            value_str = left_paren + value_str + right_paren

        segment_selector_str = "" if self.segment_selector is None else self.segment_selector

        if segment_selector_str and prefix_str:
            prefix_str += " "

            if self.offset:
                offset_str = "".join(x.render(formatting)[0] if not isinstance(x, str) else x for x in self.offset)

                # combine values and offsets according to self.offset_location
                if self.offset_location == "prefix":
                    value_str = "".join([offset_str, left_paren, value_str, right_paren])
                elif self.offset_location == "before_value":
                    value_str = "".join([left_paren, offset_str, value_str, right_paren])
                else:  # after_value
                    value_str = "".join([left_paren, value_str, offset_str, right_paren])
            else:
                value_str = left_paren + value_str + right_paren

            segment_selector_str = "" if self.segment_selector is None else self.segment_selector

            if segment_selector_str and prefix_str:
                prefix_str += " "

        return [f"{prefix_str}{segment_selector_str}{value_str}{self.suffix_str}"]


class OperandPiece(DisassemblyPiece):  # pylint: disable=abstract-method
    # These get filled in later...
    addr = None
    parentop = None
    ident = None


class Register(OperandPiece):
    def __init__(self, reg, prefix=""):
        self.reg = reg
        self.prefix = prefix
        self.is_ip = self.reg in {"eip", "rip", "pc"}  # TODO: Support more architectures

    def _render(self, formatting):
        # TODO: register renaming
        return [self.prefix + self.reg]

    def __eq__(self, other):
        return type(other) is Register and self.reg == other.reg


class Value(OperandPiece):
    def __init__(self, val, render_with_sign, render_with_pound_sign: bool = False):
        self.val = val
        self.render_with_sign = render_with_sign
        self.render_with_pound_sign = render_with_pound_sign

    @property
    def project(self):
        assert self.parentop is not None
        return self.parentop.parentinsn.project

    def __eq__(self, other):
        return type(other) is Value and self.val == other.val

    def _render(self, formatting):
        if formatting is not None:
            try:
                style = formatting["int_styles"][self.ident]
                if style[0] == "label":
                    labeloffset = style[1]
                    if labeloffset == 0:
                        lbl = self.project.kb.labels[self.val]
                        return [lbl]
                    s = "{}{}{:#+x}".format(
                        "+" if self.render_with_sign else "",
                        self.project.kb.labels[self.val + labeloffset],
                        labeloffset,
                    )
                elif style[0] == "hex":
                    s = f"{self.val:+#x}" if self.render_with_sign else f"{self.val:#x}"
                    if self.render_with_pound_sign:
                        s = "#" + s
                else:  # "dec"
                    s = f"{self.val:+d}" if self.render_with_sign else str(self.val)

                if self.render_with_pound_sign:
                    s = "#" + s
                return [s]
            except KeyError:
                pass

        # default case
        try:
            func = self.project.kb.functions.get_by_addr(self.val)
        except KeyError:
            func = None

        if self.val in self.project.kb.labels:
            lbl = self.project.kb.labels[self.val]
            if func is not None and lbl == func.name and func.name != func.demangled_name:
                # see if lbl == func.name and func.demangled_name != func.name. if so, we prioritize the
                # demangled name
                normalized_name = get_cpp_function_name(func.demangled_name)
                return [normalized_name]
            return [("+" if self.render_with_sign else "") + lbl]
        if func is not None:
            return [func.demangled_name]
        s = f"{self.val:+#x}" if self.render_with_sign else f"{self.val:#x}"
        if self.render_with_pound_sign:
            s = "#" + s
        return [s]


class Comment(DisassemblyPiece):
    def __init__(self, addr, text):
        self.addr = addr
        self.text = text.split("\n")

    def _render(self, formatting=None):
        return [self.text]

    def height(self, formatting):
        lines = len(self.text)
        return 0 if lines == 1 else lines


class FuncComment(DisassemblyPiece):
    def __init__(self, func):
        self.func = func

    def _render(self, formatting=None):
        return ["##", "## Function " + self.func.name, "##"]


class Disassembly(Analysis):
    """
    Produce formatted machine code disassembly.
    """

    def __init__(
        self,
        function: Function | None = None,
        ranges: Sequence[tuple[int, int]] | None = None,
        thumb: bool = False,
        include_ir: bool = False,
        block_bytes: bytes | None = None,
    ):
        self.raw_result = []
        self.raw_result_map = {
            "block_starts": {},
            "comments": {},
            "labels": {},
            "instructions": {},
            "hooks": {},
            "ir": defaultdict(list),
        }
        self.block_to_insn_addrs = defaultdict(list)
        self._func_cache = {}
        self._include_ir = include_ir
        self._block_bytes = block_bytes
        self._graph = None

        if function is not None:
            # sort them by address, put hooks before nonhooks
            self._graph = function.graph
            blocks = sorted(function.graph.nodes(), key=lambda node: (node.addr, not node.is_hook))
            for block in blocks:
                self.parse_block(block)
        elif ranges is not None:
            cfg = self.project.kb.cfgs.get_most_accurate()
            fallback = True
            if self._block_bytes is None and cfg is not None:
                try:
                    self._graph = cfg.graph
                    for start, end in ranges:
                        if start == end:
                            continue
                        assert start < end

                        # Grab all blocks that intersect target range
                        blocks = sorted(
                            [
                                n.to_codenode()
                                for n in self._graph.nodes()
                                if not (n.addr + (n.size or 1) <= start or n.addr >= end)
                            ],
                            key=lambda node: (node.addr, not node.is_hook),
                        )

                        # Trim blocks that are not within range
                        for i, block in enumerate(blocks):
                            if block.size and block.addr < start:
                                delta = start - block.addr
                                block_bytes = block.bytestr[delta:] if block.bytestr else None
                                blocks[i] = BlockNode(block.addr + delta, block.size - delta, block_bytes)
                        for i, block in enumerate(blocks):
                            real_block_addr = block.addr if not block.thumb else block.addr - 1
                            if block.size and real_block_addr + block.size > end:
                                delta = real_block_addr + block.size - end
                                block_bytes = block.bytestr[0:-delta] if block.bytestr else None
                                blocks[i] = BlockNode(block.addr, block.size - delta, block_bytes)

                        for block in blocks:
                            self.parse_block(block)
                    fallback = False
                except KeyError:
                    pass

            if fallback:
                # CFG not available, or the block cannot be found in the CFG (e.g., the block is dynamically
                # generated). Simply disassemble the code in the given regions. In the future we may want to handle
                # this case by automatically running CFG analysis on given ranges.
                for start, end in ranges:
                    self.parse_block(
                        BlockNode(
                            start,
                            end - start,
                            thumb=thumb,
                            bytestr=self._block_bytes if len(ranges) == 1 else None,
                        )
                    )

    def func_lookup(self, block):
        try:
            return self._func_cache[block.function.addr]
        except AttributeError:
            return None
        except KeyError:
            f = FunctionStart(block.function)
            self._func_cache[f.addr] = f
            return f

    def _add_instruction_to_results(self, block: BlockNode, insn: DisassemblerInsn, bs: BlockStart) -> None:
        """
        Add instruction to analysis results with associated labels and comments
        """
        if insn.address in self.kb.labels:
            label = Label(insn.address, self.kb.labels[insn.address])
            self.raw_result.append(label)
            self.raw_result_map["labels"][label.addr] = label
        if insn.address in self.kb.comments:
            comment = Comment(insn.address, self.kb.comments[insn.address])
            self.raw_result.append(comment)
            self.raw_result_map["comments"][comment.addr] = comment
        instruction = Instruction(insn, bs)
        self.raw_result.append(instruction)
        self.raw_result_map["instructions"][instruction.addr] = instruction
        self.block_to_insn_addrs[block.addr].append(insn.address)

    def _add_block_ir_to_results(self, block: BlockNode, irsb: IRSBType) -> None:
        """
        Add lifter IR for this block
        """
        addr_to_ops_map = self.raw_result_map["ir"]
        addr = block.addr
        ops = addr_to_ops_map[addr]

        if irsb.statements is not None:
            if pcode is not None and isinstance(self.project.factory.default_engine, pcode.HeavyPcodeMixin):
                addr = None
                for stmt_idx, op in enumerate(irsb._ops):  # type: ignore
                    if op.opcode == pypcode.OpCode.IMARK:
                        addr = op.inputs[0].offset
                    else:
                        assert addr is not None
                        addr_to_ops_map[addr].append(IROp(addr, stmt_idx, op, irsb))
            else:
                for seq, stmt in enumerate(irsb.statements):
                    if isinstance(stmt, pyvex.stmt.IMark):
                        addr = stmt.addr
                        ops = addr_to_ops_map[addr]
                    else:
                        ops.append(IROp(addr, seq, stmt, irsb))

    def parse_block(self, block: BlockNode) -> None:
        """
        Parse instructions for a given block node
        """
        func = self.func_lookup(block)
        if func and func.addr == block.addr:
            self.raw_result.append(FuncComment(block.function))
            self.raw_result.append(func)
        bs = BlockStart(block, func, self.project)
        self.raw_result.append(bs)

        if block.is_hook:
            hook = Hook(block)
            self.raw_result.append(hook)
            self.raw_result_map["hooks"][block.addr] = hook
        elif self.project.arch.capstone_support:
            # Prefer Capstone first, where we are able to extract a bit more
            # about the operands
            if block.thumb:
                aligned_block_addr = (block.addr >> 1) << 1
                cs = self.project.arch.capstone_thumb
            else:
                aligned_block_addr = block.addr
                cs = self.project.arch.capstone
            if block.bytestr is None:
                bytestr = self.project.factory.block(aligned_block_addr, block.size).bytes
            else:
                bytestr = block.bytestr
            self.block_to_insn_addrs[block.addr] = []
            for cs_insn in cs.disasm(bytestr, block.addr):
                self._add_instruction_to_results(block, CapstoneInsn(cs_insn), bs)
        elif pcode is not None and isinstance(self.project.factory.default_engine, pcode.HeavyPcodeMixin):
            # When using the P-code engine, we can fall back on its disassembly
            # in the event that Capstone does not support it
            self.block_to_insn_addrs[block.addr] = []
            b = self.project.factory.block(block.addr, size=block.size)
            for insn in b.disassembly.insns:
                self._add_instruction_to_results(block, insn, bs)
        elif type(block) is SootBlockNode:
            for raw_stmt in block.stmts:
                stmt = SootStatement(block.addr, raw_stmt)
                self.raw_result.append(stmt)
                self.raw_result_map["instructions"][stmt.addr] = stmt
                self.block_to_insn_addrs[block.addr].append(stmt.addr)
        else:
            raise AngrTypeError(
                f"Cannot disassemble block with architecture {self.project.arch} for block type {type(block)}"
            )

        if self._include_ir:
            b = self.project.factory.block(block.addr, size=block.size)
            self._add_block_ir_to_results(block, b.vex)

    def render(
        self,
        formatting=None,
        show_edges: bool = True,
        show_addresses: bool = True,
        show_bytes: bool = False,
        ascii_only: bool | None = None,
        color: bool = True,
        min_edge_depth: int = 0,
    ) -> str:
        """
        Render the disassembly to a string, with optional edges and addresses.

        Color will be added by default, if enabled. To disable color pass an empty formatting dict.
        """
        max_bytes_per_line = 5
        bytes_width = max_bytes_per_line * 3 + 1
        a2ln = defaultdict(list)
        buf = []

        if formatting is None:
            colors = (
                {
                    "address": "gray",
                    "bytes": "cyan",
                    "edge": "yellow",
                    Label: "bright_yellow",
                    ConstantOperand: "cyan",
                    MemoryOperand: "yellow",
                    Comment: "gray",
                    Hook: "green",
                }
                if ansi_color_enabled and color
                else {}
            )
            formatting = {
                "colors": colors,
                "format_callback": lambda item, s: ansi_color(s, colors.get(type(item))),
            }

        def col(item: Any) -> str | None:
            try:
                return formatting["colors"][item]
            except KeyError:
                return None

        def format_address(addr: int, color: bool = True) -> str:
            if not show_addresses:
                return ""
            a, pad = f"{addr:x}", "  "
            return (ansi_color(a, col("address")) if color else a) + pad

        def format_bytes(data: bytes, color: bool = True) -> str:
            s = " ".join(f"{x:02x}" for x in data).ljust(bytes_width)
            return ansi_color(s, col("bytes")) if color else s

        def format_comment(text: str, color: bool = True) -> str:
            s = " ; " + text
            return ansi_color(s, col(Comment)) if color else s

        comment = None

        for item in self.raw_result:
            if isinstance(item, BlockStart):
                if len(buf) > 0:
                    buf.append("")
            elif isinstance(item, Label):
                pad = len(format_address(item.addr, False)) * " "
                if show_bytes:
                    pad += bytes_width * " "
                buf.append(pad + item.render(formatting)[0])
            elif isinstance(item, Comment):
                comment = item
            elif isinstance(item, Hook):
                a2ln[item.addr].append(len(buf))
                buf.append(format_address(item.addr) + item.render(formatting)[0])
            elif isinstance(item, Instruction):
                a2ln[item.addr].append(len(buf))
                lines = []

                # Chop instruction bytes into line segments
                p, insn_bytes = 0, []
                while show_bytes and p < len(item.insn.bytes):
                    s = item.insn.bytes[p : p + min(len(item.insn.bytes) - p, max_bytes_per_line)]
                    p += len(s)
                    insn_bytes.append(s)

                # Format the instruction's address, bytes, disassembly, and comment
                s_plain = format_address(item.addr, False)
                s = format_address(item.addr)
                bytes_column = 0
                if show_bytes:
                    bytes_column = len(s_plain)
                    s_plain += format_bytes(insn_bytes[0], False)
                    s += format_bytes(insn_bytes[0])
                s_plain += item.render()[0]
                s += item.render(formatting)[0]
                comment_column = 0
                if comment is not None:
                    comment_column = len(s_plain)
                    s += format_comment(comment.text[0])
                lines.append(s)

                # Add additional lines of instruction bytes
                for i in range(1, len(insn_bytes)):
                    lines.append(" " * bytes_column + format_bytes(insn_bytes[i]))

                # Add additional lines of comments
                if comment is not None:
                    for i in range(1, len(comment.text)):
                        if len(lines) <= i:
                            lines.append(" " * comment_column)
                        lines[i] += format_comment(comment.text[i])
                    comment = None

                buf.extend(lines)
            else:
                buf.append("".join(item.render(formatting)))

        if self._graph is not None and show_edges and buf:
            edges_by_line = set()
            for edge in self._graph.edges.items():
                from_block, to_block = edge[0]
                if from_block.size is None:
                    continue
                if to_block.addr != from_block.addr + from_block.size:
                    from_addr = edge[1]["ins_addr"]
                    to_addr = to_block.addr
                    if not (from_addr in a2ln and to_addr in a2ln):
                        continue
                    for f in a2ln[from_addr]:
                        for t in a2ln[to_addr]:
                            edges_by_line.add((f, t))

            # Render block edges, to a reference buffer for tracking and output buffer for display
            edge_buf = ["" for _ in buf]
            ref_buf = ["" for _ in buf]
            edge_col = col("edge")
            for f, t in sorted(edges_by_line, key=lambda e: abs(e[0] - e[1])):
                add_edge_to_buffer(edge_buf, ref_buf, f, t, lambda s: ansi_color(s, edge_col), ascii_only=ascii_only)
                add_edge_to_buffer(ref_buf, ref_buf, f, t, ascii_only=ascii_only)
            max_edge_depth = max(*map(len, ref_buf), min_edge_depth)

            # Justify edge and combine with disassembly
            for i, line in enumerate(buf):
                buf[i] = " " * (max_edge_depth - len(ref_buf[i])) + edge_buf[i] + line

        return "\n".join(buf)


AnalysesHub.register_default("Disassembly", Disassembly)



from __future__ import annotations

DUMMY_SYMBOLIC_READ_VALUE = 0xC0DEB4BE

from .file import SimFile
from .memory_object import SimMemoryObject
from .memory_mixins import DefaultMemory

__all__ = (
    "DUMMY_SYMBOLIC_READ_VALUE",
    "DefaultMemory",
    "SimFile",
    "SimMemoryObject",
)



from __future__ import annotations
from typing import TYPE_CHECKING
from collections.abc import Callable

from .flirt_sig import FlirtSignatureParsed
from .flirt_node import FlirtNode
from .flirt_module import FlirtModule

if TYPE_CHECKING:
    from angr.knowledge_plugins.functions import Function
    from .flirt_function import FlirtFunction


# crc 16 with pre-computed table
crc16_table = [
    0x0000,
    0x1189,
    0x2312,
    0x329B,
    0x4624,
    0x57AD,
    0x6536,
    0x74BF,
    0x8C48,
    0x9DC1,
    0xAF5A,
    0xBED3,
    0xCA6C,
    0xDBE5,
    0xE97E,
    0xF8F7,
    0x1081,
    0x0108,
    0x3393,
    0x221A,
    0x56A5,
    0x472C,
    0x75B7,
    0x643E,
    0x9CC9,
    0x8D40,
    0xBFDB,
    0xAE52,
    0xDAED,
    0xCB64,
    0xF9FF,
    0xE876,
    0x2102,
    0x308B,
    0x0210,
    0x1399,
    0x6726,
    0x76AF,
    0x4434,
    0x55BD,
    0xAD4A,
    0xBCC3,
    0x8E58,
    0x9FD1,
    0xEB6E,
    0xFAE7,
    0xC87C,
    0xD9F5,
    0x3183,
    0x200A,
    0x1291,
    0x0318,
    0x77A7,
    0x662E,
    0x54B5,
    0x453C,
    0xBDCB,
    0xAC42,
    0x9ED9,
    0x8F50,
    0xFBEF,
    0xEA66,
    0xD8FD,
    0xC974,
    0x4204,
    0x538D,
    0x6116,
    0x709F,
    0x0420,
    0x15A9,
    0x2732,
    0x36BB,
    0xCE4C,
    0xDFC5,
    0xED5E,
    0xFCD7,
    0x8868,
    0x99E1,
    0xAB7A,
    0xBAF3,
    0x5285,
    0x430C,
    0x7197,
    0x601E,
    0x14A1,
    0x0528,
    0x37B3,
    0x263A,
    0xDECD,
    0xCF44,
    0xFDDF,
    0xEC56,
    0x98E9,
    0x8960,
    0xBBFB,
    0xAA72,
    0x6306,
    0x728F,
    0x4014,
    0x519D,
    0x2522,
    0x34AB,
    0x0630,
    0x17B9,
    0xEF4E,
    0xFEC7,
    0xCC5C,
    0xDDD5,
    0xA96A,
    0xB8E3,
    0x8A78,
    0x9BF1,
    0x7387,
    0x620E,
    0x5095,
    0x411C,
    0x35A3,
    0x242A,
    0x16B1,
    0x0738,
    0xFFCF,
    0xEE46,
    0xDCDD,
    0xCD54,
    0xB9EB,
    0xA862,
    0x9AF9,
    0x8B70,
    0x8408,
    0x9581,
    0xA71A,
    0xB693,
    0xC22C,
    0xD3A5,
    0xE13E,
    0xF0B7,
    0x0840,
    0x19C9,
    0x2B52,
    0x3ADB,
    0x4E64,
    0x5FED,
    0x6D76,
    0x7CFF,
    0x9489,
    0x8500,
    0xB79B,
    0xA612,
    0xD2AD,
    0xC324,
    0xF1BF,
    0xE036,
    0x18C1,
    0x0948,
    0x3BD3,
    0x2A5A,
    0x5EE5,
    0x4F6C,
    0x7DF7,
    0x6C7E,
    0xA50A,
    0xB483,
    0x8618,
    0x9791,
    0xE32E,
    0xF2A7,
    0xC03C,
    0xD1B5,
    0x2942,
    0x38CB,
    0x0A50,
    0x1BD9,
    0x6F66,
    0x7EEF,
    0x4C74,
    0x5DFD,
    0xB58B,
    0xA402,
    0x9699,
    0x8710,
    0xF3AF,
    0xE226,
    0xD0BD,
    0xC134,
    0x39C3,
    0x284A,
    0x1AD1,
    0x0B58,
    0x7FE7,
    0x6E6E,
    0x5CF5,
    0x4D7C,
    0xC60C,
    0xD785,
    0xE51E,
    0xF497,
    0x8028,
    0x91A1,
    0xA33A,
    0xB2B3,
    0x4A44,
    0x5BCD,
    0x6956,
    0x78DF,
    0x0C60,
    0x1DE9,
    0x2F72,
    0x3EFB,
    0xD68D,
    0xC704,
    0xF59F,
    0xE416,
    0x90A9,
    0x8120,
    0xB3BB,
    0xA232,
    0x5AC5,
    0x4B4C,
    0x79D7,
    0x685E,
    0x1CE1,
    0x0D68,
    0x3FF3,
    0x2E7A,
    0xE70E,
    0xF687,
    0xC41C,
    0xD595,
    0xA12A,
    0xB0A3,
    0x8238,
    0x93B1,
    0x6B46,
    0x7ACF,
    0x4854,
    0x59DD,
    0x2D62,
    0x3CEB,
    0x0E70,
    0x1FF9,
    0xF78F,
    0xE606,
    0xD49D,
    0xC514,
    0xB1AB,
    0xA022,
    0x92B9,
    0x8330,
    0x7BC7,
    0x6A4E,
    0x58D5,
    0x495C,
    0x3DE3,
    0x2C6A,
    0x1EF1,
    0x0F78,
]


def crc16(data: bytes) -> int:
    crc = 0xFFFF
    for byte in data:
        crc = (crc >> 8) ^ crc16_table[(crc ^ byte) & 0xFF]
    crc ^= 0xFFFF
    # swap endianness
    return ((crc & 0xFF) << 8) | ((crc & 0xFF00) >> 8)


class FlirtMatcher:
    """
    A class that matches functions in a binary using FLIRT signatures.
    """

    def __init__(
        self,
        sig: FlirtSignatureParsed,
        func: Function,
        get_callee_name: Callable[
            [Function, int, int, str],
            str | None,
        ],
        func_matched: Callable[[Function, int, FlirtFunction], None],
        mismatch_bytes_tolerance: int = 0,
    ):
        self.sig = sig
        self.func = func
        self.get_callee_name = get_callee_name
        self.func_matched = func_matched
        self.mismatch_bytes_tolerance: int = mismatch_bytes_tolerance

    def match_function(self, buff: bytes, addr: int) -> bool:
        assert self.sig.root is not None
        return any(self._match_node(node, buff, addr, 0, 0) for node in self.sig.root.children)

    def _match_node(self, node: FlirtNode, buff: bytes, addr: int, offset: int, mismatches: int) -> bool:
        if len(buff) < offset + len(node.pattern):
            return False
        for i in range(len(node.pattern)):  # pylint:disable=consider-using-enumerate
            if node.pattern[i] != -1 and node.pattern[i] != buff[offset + i]:
                mismatches += 1
                if mismatches > self.mismatch_bytes_tolerance:
                    return False
        if mismatches <= self.mismatch_bytes_tolerance:
            # a matching node is found
            for child in node.children:
                if self._match_node(child, buff, addr, offset + node.length, mismatches):
                    return True
            for module in node.modules:
                if self._match_module(module, buff, addr, offset + node.length):
                    return True
        return False

    def _match_module(self, module: FlirtModule, buff: bytes, addr: int, offset: int) -> bool:
        offset = max(offset, 32)
        if module.crc_len > len(buff) - offset:
            return False
        crc = crc16(buff[offset : offset + module.crc_len]) if module.crc_len > 0 else 0
        if crc != module.crc:
            return False

        # tail bytes
        for off, b in module.tail_bytes:
            tail_byte_buff_off = offset + module.crc_len + off
            if len(buff) <= tail_byte_buff_off or buff[tail_byte_buff_off] != b:
                return False

        # referenced functions
        for ref_func in module.ref_funcs:
            call_addr = addr + offset + ref_func.offset
            callee_name = self.get_callee_name(self.func, addr, call_addr, ref_func.name)
            if callee_name != ref_func.name:
                return False

        for func in module.pub_funcs:
            self.func_matched(self.func, addr, func)

        return True



from __future__ import annotations
from typing import Any, TYPE_CHECKING
import os
import logging
import tempfile
import uuid
import contextlib
import shutil
import weakref
from collections import defaultdict

import lmdb

from angr.knowledge_plugins.plugin import KnowledgeBasePlugin
from angr.errors import AngrRuntimeDbError

if TYPE_CHECKING:
    from angr.knowledge_base import KnowledgeBase

RTDB_BASEDIR: str | None = os.environ.get("RTDB_BASE")


l = logging.getLogger(__name__)


class RuntimeDbForkCondom:
    """
    A class that invokes RuntimeDb.reopen_lmdb() upon forking. This is necessary to ensure that lmdb does not raise
    ReaderFullError in forked child processes. The reopen_rtdb() method is not called on Windows because
    os.register_at_fork does not exist on Windows.
    """

    def __init__(self, rtdb: RuntimeDb):
        self.rtdb = weakref.proxy(rtdb)
        if hasattr(os, "register_at_fork"):
            os.register_at_fork(after_in_child=self.reopen_rtdb)

    def reopen_rtdb(self):
        if self.rtdb is None:
            return
        try:
            self.rtdb.reopen_lmdb()
        except ReferenceError:
            self.rtdb = None


class RuntimeDb(KnowledgeBasePlugin):
    """
    External storage-backed database for angr knowledge base plugins.
    """

    def __init__(self, kb: KnowledgeBase, lmdb_path: str | None = None) -> None:
        super().__init__(kb)

        self._lmdb_path: str | None = lmdb_path
        self._lmdb_env: lmdb.Environment | None = None
        self._lmdb_mapsize: int = 1024 * 1024 * 10
        self._dbnames: defaultdict[str, int] = defaultdict(int)
        self._dbs: dict[str, Any] = {}
        self._condom = RuntimeDbForkCondom(self)

    def __del__(self):
        self.cleanup()

    def _init_lmdb(self):
        if self._lmdb_env is not None:
            return

        if self._lmdb_path is not None:
            # may fail and raise exceptions
            self._lmdb_env = self._attempt_creating_lmdb(self._lmdb_path)
        else:
            r = self._init_lmdb_attempt_multipele_locations()
            if r is None:
                raise AngrRuntimeDbError("Failed to initialize the LMDB environment.")
            self._lmdb_path, self._lmdb_env = r

        l.debug("Initialized LRU cache LMDB at %s", self._lmdb_path)

    def _init_lmdb_attempt_multipele_locations(self) -> tuple[str, lmdb.Environment] | None:
        if self._lmdb_env is not None:
            return None

        main_binary_path = self._kb._project.loader.main_object.binary
        basename = os.path.basename(main_binary_path) if isinstance(main_binary_path, str) else "angr_proj"

        basedir = None
        if RTDB_BASEDIR is not None:
            basedir = RTDB_BASEDIR
            if not os.access(basedir, os.W_OK):
                l.error("The directory %s is not writable. Falling back.", basedir)
                basedir = None
            else:
                db_filename = self._get_unique_db_filename(basedir, basename)
                lmdb_path = os.path.join(basedir, db_filename)
                lmdb_env = self._attempt_creating_lmdb(lmdb_path)
                if lmdb_env is not None:
                    return lmdb_path, lmdb_env
                basedir = None
                db_filename = None

        if basedir is None and isinstance(main_binary_path, str):
            # get the base directory of the project binary
            basedir = os.path.dirname(os.path.abspath(main_binary_path))
            # is the location writable?
            if not os.access(basedir, os.W_OK):
                l.error("The directory %s is not writable. Falling back to temporary directory.", basedir)
                basedir = None
            else:
                db_filename = self._get_unique_db_filename(basedir, basename)
                lmdb_path = os.path.join(basedir, db_filename)
                lmdb_env = self._attempt_creating_lmdb(lmdb_path)
                if lmdb_env is not None:
                    return lmdb_path, lmdb_env
                basedir = None
                db_filename = None

        if basedir is None:
            basedir = tempfile.gettempdir()
            if not os.access(basedir, os.W_OK):
                raise OSError("No writable directory found for RTDB storage.")
            db_filename = self._get_unique_db_filename(basedir, basename)
            lmdb_path = os.path.join(basedir, db_filename)
            lmdb_env = self._attempt_creating_lmdb(lmdb_path)
            if lmdb_env is not None:
                return lmdb_path, lmdb_env
            raise OSError("No writable directory found for RTDB storage.")

        return None

    def _attempt_creating_lmdb(self, lmdb_path: str):
        """
        Attempt to create the LMDB environment.
        """
        if self._lmdb_env is not None:
            return None

        try:
            return lmdb.open(lmdb_path, sync=False, map_size=self._lmdb_mapsize, max_dbs=10)
        except (PermissionError, OSError):
            return None

    @staticmethod
    def _get_unique_db_filename(basedir: str, basename: str) -> str:
        # find a unique rtdb name
        db_filename = basename + "_angr_rtdb"
        while True:
            db_path = os.path.join(basedir, db_filename)
            if not os.path.exists(db_path):
                break
            db_filename = basename + f"_angr_rtdb_{uuid.uuid4().hex}"
        return db_filename

    def _cleanup_lmdb(self):
        """
        Clean up LMDB resources.
        """
        if self._lmdb_env is not None:
            self._lmdb_env.close()
            self._lmdb_env = None

        if self._lmdb_path is not None:
            with contextlib.suppress(OSError):
                shutil.rmtree(self._lmdb_path)
            self._lmdb_path = None

    def _get_unique_dbname(self, base_name: str) -> str:
        count = self._dbnames[base_name]
        self._dbnames[base_name] += 1
        if count == 0:
            return base_name
        return f"{base_name}_{count}"

    def increase_lmdb_map_size(self) -> None:
        """
        Increase the LMDB map size.
        """
        if self._lmdb_env is None:
            return

        delta = min(self._lmdb_mapsize, 1024 * 1024 * 256)
        l.debug("Increasing LMDB map size by %d bytes", delta)
        self._lmdb_mapsize += delta
        self._lmdb_env.set_mapsize(self._lmdb_mapsize)

    def reopen_lmdb(self):
        """
        Reopen the existing LMDB environment and all open databases in self._dbs.
        """

        if self._lmdb_path is None:
            # we've never opened any LMDB before
            return

        if self._lmdb_env is not None:
            self._lmdb_env.close()
            self._lmdb_env = None

        self._init_lmdb()
        assert self._lmdb_env is not None

        for db_name in list(self._dbs):
            self._dbs[db_name] = self._lmdb_env.open_db(db_name.encode())

    def open_db(self, db_name: str, unique: bool = True) -> str:
        self._init_lmdb()
        assert self._lmdb_env is not None
        if unique:
            db_name = self._get_unique_dbname(db_name)
        if db_name in self._dbs:
            return db_name
        db = self._lmdb_env.open_db(db_name.encode())
        self._dbs[db_name] = db
        return db_name

    def begin_txn(self, db_name: str, write: bool = False):
        db = self._dbs[db_name]
        assert self._lmdb_env is not None
        return self._lmdb_env.begin(db=db, write=write)

    def drop_db(self, db_name: str) -> None:
        db = self._dbs[db_name]
        if self._lmdb_env is None:
            return
        with self._lmdb_env.begin(write=True) as txn:
            txn.drop(db)
        del self._dbs[db_name]

    def cleanup(self):
        self._cleanup_lmdb()


KnowledgeBasePlugin.register_default("rtdb", RuntimeDb)



# pylint:disable=unused-argument,no-self-use
from __future__ import annotations
from typing import TYPE_CHECKING
from itertools import count

from angr import sim_type
from angr.sim_type import SimType
from . import typeconsts
from .typeconsts import TypeConstant

if TYPE_CHECKING:
    import archinfo


class SimTypeTempRef(sim_type.SimType):
    """
    Represents a temporary reference to another type. TypeVariableReference is translated to SimTypeTempRef.
    """

    def __init__(self, typevar):
        super().__init__()
        self.typevar = typevar

    def c_repr(self, **kwargs):
        return "<SimTypeTempRef>"


class TypeTranslator:
    """
    Translate type constants to SimType and SimType to type constants.
    """

    __slots__ = (
        "_has_nonexistent_ref",
        "_struct_ctr",
        "arch",
        "memo",
        "named_struct_id_counter",
        "struct_name_to_idx",
        "structs",
        "translated",
        "translated_simtypes",
    )

    def __init__(self, arch: archinfo.Arch):
        self.arch: archinfo.Arch = arch

        self.translated: dict[TypeConstant, SimType] = {}
        self.translated_simtypes: dict[SimType, TypeConstant] = {}
        self.structs = {}
        self._struct_ctr = count()
        self.memo = {}
        self.named_struct_id_counter = count(133337)
        self.struct_name_to_idx = {}

        # will be updated every time .tc2simtype() is called
        self._has_nonexistent_ref = False

    #
    # Naming
    #

    def struct_name(self):
        return f"struct_{next(self._struct_ctr)}"

    #
    # Type translation
    #

    def tc2simtype(self, tc):
        self._has_nonexistent_ref = False
        return self._tc2simtype(tc), self._has_nonexistent_ref

    def _tc2simtype(self, tc):
        if tc is None:
            return sim_type.SimTypeBottom().with_arch(self.arch)

        try:
            handler = TypeConstHandlers[tc.__class__]
        except KeyError:
            return sim_type.SimTypeBottom().with_arch(self.arch)

        return handler(self, tc)

    def simtype2tc(self, simtype: sim_type.SimType) -> typeconsts.TypeConstant:
        return self._simtype2tc(simtype)

    def _simtype2tc(self, simtype: sim_type.SimType) -> typeconsts.TypeConstant:
        if simtype in self.translated_simtypes:
            return self.translated_simtypes[simtype]
        try:
            handler = SimTypeHandlers[simtype.__class__]
            return handler(self, simtype)
        except KeyError:
            return typeconsts.BottomType()

    # compatibility and quality of life
    lift = simtype2tc
    concretize = tc2simtype

    #
    # Typehoon type handlers
    #

    def _translate_Pointer64(self, tc):
        if isinstance(tc.basetype, typeconsts.BottomType):
            # void *
            internal = sim_type.SimTypeBottom(label="void").with_arch(self.arch)
        else:
            internal = self._tc2simtype(tc.basetype)
        return sim_type.SimTypePointer(internal, label=tc.name).with_arch(self.arch)

    def _translate_Pointer32(self, tc):
        if isinstance(tc.basetype, typeconsts.BottomType):
            # void *
            internal = sim_type.SimTypeBottom(label="void").with_arch(self.arch)
        else:
            internal = self._tc2simtype(tc.basetype)
        return sim_type.SimTypePointer(internal, label=tc.name).with_arch(self.arch)

    def _translate_Array(self, tc: typeconsts.Array) -> sim_type.SimTypeArray:
        elem_type = self._tc2simtype(tc.element)
        return sim_type.SimTypeArray(elem_type, length=tc.count, label=tc.name).with_arch(self.arch)

    def _translate_Struct(self, tc: typeconsts.Struct):
        if tc in self.structs:
            return self.structs[tc]

        name = tc.name or self.struct_name()

        if tc.is_cppclass:
            s = sim_type.SimCppClass(name=name).with_arch(self.arch)
        else:
            s = sim_type.SimStruct({}, name=name).with_arch(self.arch)
        self.structs[tc] = s

        next_offset = 0
        for offset, typ in sorted(tc.fields.items(), key=lambda item: item[0]):
            if offset > next_offset:
                # we need padding!
                padding_size = offset - next_offset
                s.fields[f"padding_{next_offset:x}"] = sim_type.SimTypeFixedSizeArray(
                    sim_type.SimTypeChar(signed=False).with_arch(self.arch), padding_size
                ).with_arch(self.arch)

            translated_type = self._tc2simtype(typ)
            if isinstance(translated_type, sim_type.SimTypeBottom):
                # we cannot have bottom types in a struct since that would mess with offsets of all future types
                # for now, we replace it with an unsigned char
                translated_type = sim_type.SimTypeChar(signed=False).with_arch(self.arch)

            field_name = tc.field_names[offset] if tc.field_names and offset in tc.field_names else f"field_{offset:x}"
            s.fields[field_name] = translated_type

            if isinstance(translated_type, SimTypeTempRef):
                next_offset = self.arch.bytes + offset
            else:
                next_offset = translated_type.size // self.arch.byte_width + offset

        return s

    def _translate_Int8(self, tc):
        return sim_type.SimTypeChar(signed=False, label=tc.name).with_arch(self.arch)

    def _translate_Int16(self, tc):
        if tc.name == "WCHAR":
            return sim_type.SimTypeWideChar(label=tc.name).with_arch(self.arch)
        return sim_type.SimTypeShort(signed=False, label=tc.name).with_arch(self.arch)

    def _translate_Int32(self, tc):
        return sim_type.SimTypeInt(signed=False, label=tc.name).with_arch(self.arch)

    def _translate_Int64(self, tc):
        return sim_type.SimTypeLongLong(signed=False, label=tc.name).with_arch(self.arch)

    def _translate_Int128(self, tc):
        return sim_type.SimTypeInt128(signed=False, label=tc.name).with_arch(self.arch)

    def _translate_Int256(self, tc):
        return sim_type.SimTypeInt256(signed=False, label=tc.name).with_arch(self.arch)

    def _translate_Int512(self, tc):
        return sim_type.SimTypeInt512(signed=False, label=tc.name).with_arch(self.arch)

    def _translate_TypeVariableReference(self, tc):
        if tc.typevar in self.translated:
            return self.translated[tc.typevar]

        self._has_nonexistent_ref = True
        return SimTypeTempRef(tc.typevar)

    def _translate_Float32(self, tc: typeconsts.Float32) -> sim_type.SimTypeFloat:
        return sim_type.SimTypeFloat(label=tc.name).with_arch(self.arch)

    def _translate_Float64(self, tc: typeconsts.Float64) -> sim_type.SimTypeDouble:
        return sim_type.SimTypeDouble(label=tc.name).with_arch(self.arch)

    def _translate_Enum(self, tc: typeconsts.Enum) -> sim_type.SimTypeEnum:
        """Convert Enum type constant to SimTypeEnum."""
        base_simtype = None
        if tc.base_type is not None:
            base_simtype = self._tc2simtype(tc.base_type)
        return sim_type.SimTypeEnum(
            members=dict(tc.members),
            name=tc.name,
            base_type=base_simtype,
        ).with_arch(self.arch)

    def _translate_Fd(self, tc: typeconsts.Fd) -> sim_type.SimTypeFd:
        return sim_type.SimTypeFd(label=tc.name).with_arch(self.arch)

    #
    # Backpatching
    #

    def backpatch(self, st, translated):
        """

        :param sim_type.SimType st:
        :param dict translated:
        :return:
        """

        if isinstance(st, sim_type.SimTypePointer):
            self.backpatch(st.pts_to, translated)

        elif isinstance(st, sim_type.SimStruct):
            fields_patch = {}
            for offset, fld in st.fields.items():
                if isinstance(fld, SimTypeTempRef) and fld.typevar in translated:
                    fields_patch[offset] = translated[fld.typevar]
                st.fields.update(fields_patch)

    #
    # SimType handlers
    #

    def _translate_SimTypeInt128(self, st: sim_type.SimTypeChar) -> typeconsts.Int128:
        return typeconsts.Int128(name=st.label)

    def _translate_SimTypeInt256(self, st: sim_type.SimTypeChar) -> typeconsts.Int256:
        return typeconsts.Int256(name=st.label)

    def _translate_SimTypeInt512(self, st: sim_type.SimTypeChar) -> typeconsts.Int512:
        return typeconsts.Int512(name=st.label)

    def _translate_SimTypeInt(self, st: sim_type.SimTypeInt) -> typeconsts.Int32:
        return typeconsts.Int32(name=st.label)

    def _translate_SimTypeLong(self, st: sim_type.SimTypeLong) -> typeconsts.Int32:
        return typeconsts.Int32(name=st.label)

    def _translate_SimTypeLongLong(self, st: sim_type.SimTypeLongLong) -> typeconsts.Int64:
        return typeconsts.Int64(name=st.label)

    def _translate_SimTypeShort(self, st: sim_type.SimTypeInt) -> typeconsts.Int16:
        return typeconsts.Int16(name=st.label)

    def _translate_SimTypeChar(self, st: sim_type.SimTypeChar) -> typeconsts.Int8:
        return typeconsts.Int8(name=st.label)

    def _translate_SimTypeWideChar(self, st: sim_type.SimTypeWideChar) -> typeconsts.Int16:
        return typeconsts.Int16(name=st.label)

    def _translate_SimStruct(self, st: sim_type.SimStruct) -> typeconsts.Struct | typeconsts.BottomType:
        if st in self.memo:
            return typeconsts.BottomType()

        struct_idx = {}
        if st.name:
            if st.name not in self.struct_name_to_idx:
                self.struct_name_to_idx[st.name] = next(self.named_struct_id_counter)
            struct_idx["idx"] = self.struct_name_to_idx[st.name]

        obj = typeconsts.Struct(fields={}, name=st.name, **struct_idx)
        self.memo[st] = obj

        fields = {}
        field_names = {}
        offsets = st.offsets
        for field_name, simtype in st.fields.items():
            if field_name not in offsets:
                return typeconsts.BottomType()
            offset = offsets[field_name]
            fields[offset] = self._simtype2tc(simtype)
            field_names[offsets[field_name]] = field_name
        obj.fields = fields
        obj.field_names = field_names
        del self.memo[st]

        return obj

    def _translate_SimCppClass(self, st: sim_type.SimCppClass) -> typeconsts.Struct | typeconsts.BottomType:
        if st in self.memo:
            return typeconsts.BottomType()

        struct_idx = {}
        if st.name:
            if st.name not in self.struct_name_to_idx:
                self.struct_name_to_idx[st.name] = next(self.named_struct_id_counter)
            struct_idx["idx"] = self.struct_name_to_idx[st.name]

        obj = typeconsts.Struct(fields={}, name=st.name, is_cppclass=True, **struct_idx)
        self.memo[st] = obj

        fields = {}
        field_names = {}
        offsets = st.offsets
        for field_name, simtype in st.fields.items():
            if field_name not in offsets:
                return typeconsts.BottomType()
            offset = offsets[field_name]
            fields[offset] = self._simtype2tc(simtype)
            field_names[offsets[field_name]] = field_name
        obj.fields = fields
        obj.field_names = field_names
        del self.memo[st]

        return obj

    def _translate_SimTypeArray(self, st: sim_type.SimTypeArray) -> typeconsts.Array:
        elem_type = self._simtype2tc(st.elem_type)
        return typeconsts.Array(elem_type, count=st.length, name=st.label)

    def _translate_SimTypePointer(self, st: sim_type.SimTypePointer) -> typeconsts.Pointer32 | typeconsts.Pointer64:
        base = self._simtype2tc(st.pts_to)
        if self.arch.bits == 32:
            return typeconsts.Pointer32(base, name=st.label)
        if self.arch.bits == 64:
            return typeconsts.Pointer64(base, name=st.label)
        raise TypeError(f"Unsupported pointer size {self.arch.bits}")

    def _translate_SimTypeFloat(self, st: sim_type.SimTypeFloat) -> typeconsts.Float32:
        return typeconsts.Float32(name=st.label)

    def _translate_SimTypeDouble(self, st: sim_type.SimTypeDouble) -> typeconsts.Float64:
        return typeconsts.Float64(name=st.label)

    def _translate_SimTypeEnum(self, st: sim_type.SimTypeEnum) -> typeconsts.Enum:
        """Convert SimTypeEnum to Enum type constant."""
        base_tc = None
        if st._base_type is not None:
            base_tc = self._simtype2tc(st._base_type)
        return typeconsts.Enum(
            members=dict(st.members),
            base_type=base_tc,
            name=st.name,
        )

    def _translate_SimTypeFd(self, st: sim_type.SimTypeFd) -> typeconsts.Fd:
        return typeconsts.Fd(name=st.label)


TypeConstHandlers = {
    typeconsts.Pointer64: TypeTranslator._translate_Pointer64,
    typeconsts.Pointer32: TypeTranslator._translate_Pointer32,
    typeconsts.Array: TypeTranslator._translate_Array,
    typeconsts.Struct: TypeTranslator._translate_Struct,
    typeconsts.Enum: TypeTranslator._translate_Enum,
    typeconsts.Int8: TypeTranslator._translate_Int8,
    typeconsts.Int16: TypeTranslator._translate_Int16,
    typeconsts.Int32: TypeTranslator._translate_Int32,
    typeconsts.Int64: TypeTranslator._translate_Int64,
    typeconsts.Int128: TypeTranslator._translate_Int128,
    typeconsts.Int256: TypeTranslator._translate_Int256,
    typeconsts.Int512: TypeTranslator._translate_Int512,
    typeconsts.TypeVariableReference: TypeTranslator._translate_TypeVariableReference,
    typeconsts.Float32: TypeTranslator._translate_Float32,
    typeconsts.Float64: TypeTranslator._translate_Float64,
    typeconsts.Fd: TypeTranslator._translate_Fd,
}


SimTypeHandlers = {
    sim_type.SimTypePointer: TypeTranslator._translate_SimTypePointer,
    sim_type.SimTypeChar: TypeTranslator._translate_SimTypeChar,
    sim_type.SimTypeWideChar: TypeTranslator._translate_SimTypeWideChar,
    sim_type.SimTypeInt: TypeTranslator._translate_SimTypeInt,
    sim_type.SimTypeShort: TypeTranslator._translate_SimTypeShort,
    sim_type.SimTypeLong: TypeTranslator._translate_SimTypeLong,
    sim_type.SimTypeLongLong: TypeTranslator._translate_SimTypeLongLong,
    sim_type.SimTypeInt128: TypeTranslator._translate_SimTypeInt128,
    sim_type.SimTypeInt256: TypeTranslator._translate_SimTypeInt256,
    sim_type.SimTypeInt512: TypeTranslator._translate_SimTypeInt512,
    sim_type.SimStruct: TypeTranslator._translate_SimStruct,
    sim_type.SimTypeArray: TypeTranslator._translate_SimTypeArray,
    sim_type.SimTypeFloat: TypeTranslator._translate_SimTypeFloat,
    sim_type.SimTypeDouble: TypeTranslator._translate_SimTypeDouble,
    sim_type.SimTypeEnum: TypeTranslator._translate_SimTypeEnum,
    sim_type.SimCppClass: TypeTranslator._translate_SimCppClass,
    sim_type.SimTypeFd: TypeTranslator._translate_SimTypeFd,
}



from __future__ import annotations
import angr

import logging

l = logging.getLogger(name=__name__)


class bind(angr.SimProcedure):
    # pylint:disable=arguments-differ

    def run(self, fd, addr_ptr, addr_len):  # pylint:disable=unused-argument
        return self.state.solver.Unconstrained("bind", self.arch.sizeof["int"], key=("api", "bind"))



from __future__ import annotations

from .data_dependency_analysis import DataDependencyGraphAnalysis
from .dep_nodes import DepNodeTypes, BaseDepNode, VarDepNode, MemDepNode, ConstantDepNode, TmpDepNode, RegDepNode

__all__ = (
    "BaseDepNode",
    "ConstantDepNode",
    "DataDependencyGraphAnalysis",
    "DepNodeTypes",
    "MemDepNode",
    "RegDepNode",
    "TmpDepNode",
    "VarDepNode",
)



from __future__ import annotations
import logging
from typing import Any, TYPE_CHECKING
from collections.abc import Callable

from angr.ailment.block import Block as AILBlock
from angr.ailment.statement import Statement
from angr.ailment.expression import Expression, BinaryOp

from angr.analyses.decompiler.sequence_walker import SequenceWalker
from angr.analyses.decompiler.ail_simplifier import AILBlockRewriter
from .optimization_pass import SequenceOptimizationPass, OptimizationPassStage

if TYPE_CHECKING:
    from angr.analyses.decompiler.structurer_nodes import ConditionNode, ConditionalBreakNode, LoopNode


_l = logging.getLogger(__name__)


class OuterWalker(SequenceWalker):
    """
    A sequence walker that finds nodes and invokes expression replacer to replace expressions.
    """

    def __init__(self, desc):
        super().__init__()
        self.desc: dict[OpDescriptor, str] = desc

    def _handle_Condition(self, node: ConditionNode, **kwargs):
        for desc, new_op in self.desc.items():
            if (
                node.condition is not None
                and "ins_addr" in node.condition.tags
                and node.condition.tags["ins_addr"] == desc.ins_addr
                and node.condition.op == desc.op
            ):
                node.condition = self._swap_expr_op(new_op, node.condition)
        return super()._handle_Condition(node, **kwargs)

    def _handle_Loop(self, node: LoopNode, **kwargs):
        for desc, new_op in self.desc.items():
            if (
                node.condition is not None
                and "ins_addr" in node.condition.tags
                and node.condition.tags["ins_addr"] == desc.ins_addr
                and node.condition.op == desc.op
            ):
                node.condition = self._swap_expr_op(new_op, node.condition)
        return super()._handle_Loop(node, **kwargs)

    def _handle_ConditionalBreak(self, node: ConditionalBreakNode, **kwargs):
        for desc, new_op in self.desc.items():
            if (
                node.condition is not None
                and "ins_addr" in node.condition.tags
                and node.condition.tags["ins_addr"] == desc.ins_addr
                and node.condition.op == desc.op
            ):
                node.condition = self._swap_expr_op(new_op, node.condition)
        return super()._handle_ConditionalBreak(node, **kwargs)

    @staticmethod
    def _swap_expr_op(new_op: str, atom: Expression) -> Expression | None:
        # swap
        return BinaryOp(
            atom.idx, new_op, (atom.operands[1], atom.operands[0]), atom.signed, bits=atom.bits, **atom.tags
        )


class ExpressionReplacer(AILBlockRewriter):
    """
    Replace expressions.
    """

    def __init__(self, block_addr, target_expr_predicate, callback):
        super().__init__()
        self._block_addr = block_addr
        self._target_expr_predicate: Callable = target_expr_predicate
        self._callback = callback

    def _handle_expr(
        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: AILBlock | None
    ) -> Any:
        if self._target_expr_predicate(expr):
            return self._callback(self._block_addr, expr)
        return super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)


class OpDescriptor:
    """
    Describes a specific operator.
    """

    def __init__(self, block_addr: int, stmt_idx: int, ins_addr: int, op: str):
        self.block_addr = block_addr
        self.stmt_idx = stmt_idx
        self.ins_addr = ins_addr
        self.op = op

    def __hash__(self):
        return hash((OpDescriptor, self.block_addr, self.stmt_idx, self.ins_addr, self.op))

    def __eq__(self, other):
        return (
            isinstance(other, OpDescriptor)
            and self.block_addr == other.block_addr
            and self.stmt_idx == other.stmt_idx
            and self.ins_addr == other.ins_addr
            and self.op == other.op
        )


class ExprOpSwapper(SequenceOptimizationPass):
    """
    Swap operands (and the operator accordingly) in a BinOp expression.
    """

    ARCHES = ["X86", "AMD64", "ARMEL", "ARMHF", "ARMCortexM", "MIPS32", "MIPS64"]
    PLATFORMS = ["windows", "linux", "cgc"]
    STAGE = OptimizationPassStage.AFTER_STRUCTURING
    NAME = "Swap operands of expressions as requested"
    DESCRIPTION = __doc__.strip()

    def __init__(self, *args, binop_operators: dict[OpDescriptor, str] | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._expr_operators = {} if binop_operators is None else binop_operators

        if self._expr_operators:
            self.analyze()

    def _check(self):
        return True, None

    def _analyze(self, cache=None):  # pylint:disable=unused-argument
        walker = OuterWalker(self._expr_operators)
        walker.walk(self.seq)
        self.out_seq = self.seq



# pylint:disable=missing-class-docstring,unused-argument,consider-using-dict-items
from __future__ import annotations
from collections import defaultdict
from collections.abc import Iterable
from typing import Any, TYPE_CHECKING

from angr import ailment
from angr.ailment import Expression, Block, AILBlockRewriter
from angr.ailment.expression import Call, ITE, Atom, Load, VirtualVariable
from angr.ailment.statement import Statement, Assignment, Return

from angr.utils.ail import is_phi_assignment
from angr.utils.ssa import VVarUsesCollector
from angr.analyses.decompiler.sequence_walker import SequenceWalker
from angr.analyses.decompiler.structuring.structurer_nodes import (
    ConditionNode,
    ConditionalBreakNode,
    LoopNode,
    CascadingConditionNode,
    SequenceNode,
    SwitchCaseNode,
)

if TYPE_CHECKING:
    from angr.ailment.expression import MultiStatementExpression


class LocationBase:
    __slots__ = ()


class StatementLocation(LocationBase):
    """
    Describes the location of a statement.
    """

    __slots__ = (
        "block_addr",
        "block_idx",
        "phi_stmt",
        "stmt_idx",
    )

    def __init__(self, block_addr, block_idx, stmt_idx, phi_stmt: bool = False):
        self.block_addr = block_addr
        self.block_idx = block_idx
        self.stmt_idx = stmt_idx
        self.phi_stmt = phi_stmt

    def __repr__(self):
        return f"Loc: Statement@{self.block_addr:x}.{self.block_idx}-{self.stmt_idx}{' phi' if self.phi_stmt else ''}"

    def __hash__(self):
        return hash((StatementLocation, self.block_addr, self.block_idx, self.stmt_idx, self.phi_stmt))

    def __eq__(self, other):
        return (
            isinstance(other, StatementLocation)
            and self.block_addr == other.block_addr
            and self.block_idx == other.block_idx
            and self.stmt_idx == other.stmt_idx
            and self.phi_stmt == other.phi_stmt
        )

    def copy(self):
        return StatementLocation(self.block_addr, self.block_idx, self.stmt_idx, phi_stmt=self.phi_stmt)


class ExpressionLocation(LocationBase):
    """
    Describes the location of an expression.
    """

    __slots__ = (
        "block_addr",
        "block_idx",
        "expr_idx",
        "phi_stmt",
        "stmt_idx",
    )

    def __init__(self, block_addr, block_idx, stmt_idx, expr_idx, phi_stmt: bool = False):
        self.block_addr = block_addr
        self.block_idx = block_idx
        self.stmt_idx = stmt_idx
        self.expr_idx = expr_idx
        self.phi_stmt = phi_stmt

    def __repr__(self):
        return (
            f"Loc: Expression@{self.block_addr:x}.{self.block_idx}-{self.stmt_idx}[{self.expr_idx}]"
            f"{'phi' if self.phi_stmt else ''}"
        )

    def statement_location(self) -> StatementLocation:
        return StatementLocation(self.block_addr, self.block_idx, self.stmt_idx, phi_stmt=self.phi_stmt)

    def __hash__(self):
        return hash((ExpressionLocation, self.block_addr, self.block_idx, self.stmt_idx, self.expr_idx, self.phi_stmt))

    def __eq__(self, other):
        return (
            isinstance(other, ExpressionLocation)
            and self.block_addr == other.block_addr
            and self.block_idx == other.block_idx
            and self.stmt_idx == other.stmt_idx
            and self.expr_idx == other.expr_idx
            and self.phi_stmt == other.phi_stmt
        )


class ConditionLocation(LocationBase):
    """
    Describes the location of a condition.
    """

    __slots__ = (
        "case_idx",
        "node_addr",
    )

    def __init__(self, cond_node_addr, case_idx: int | None = None):
        self.node_addr = cond_node_addr
        self.case_idx = case_idx

    def __repr__(self):
        return f"Loc: ConditionNode@{self.node_addr:x}.{self.case_idx}"

    def __hash__(self):
        return hash((ConditionLocation, self.node_addr, self.case_idx))

    def __eq__(self, other):
        return (
            isinstance(other, ConditionLocation)
            and self.node_addr == other.node_addr
            and self.case_idx == other.case_idx
        )


class ConditionalBreakLocation(LocationBase):
    """
    Describes the location of a conditional break.
    """

    __slots__ = ("node_addr",)

    def __init__(self, node_addr):
        self.node_addr = node_addr

    def __repr__(self):
        return f"Loc: ConditionalBreakNode@{self.node_addr:x}"

    def __hash__(self):
        return hash((ConditionalBreakLocation, self.node_addr))

    def __eq__(self, other):
        return isinstance(other, ConditionalBreakLocation) and self.node_addr == other.node_addr


class LoopNodeFinder(SequenceWalker):
    """
    Returns all loops in a given region.
    """

    def __init__(self, node: SequenceNode):
        handlers = {
            LoopNode: self._handle_Loop,
        }
        super().__init__(handlers, update_seqnode_in_place=False, force_forward_scan=True)
        self.loop_nodes: list[LoopNode] = []

        self.walk(node)

    def _handle_Loop(self, node: LoopNode, **kwargs):
        super()._handle_Loop(node, **kwargs)
        self.loop_nodes.append(node)
        return None


class MultiStatementExpressionAssignmentFinder(AILBlockRewriter):
    """
    Process statements in MultiStatementExpression objects and find assignments.
    """

    def __init__(self, stmt_handler):
        super().__init__()
        self._stmt_handler = stmt_handler

    def _handle_MultiStatementExpression(
        self, expr_idx, expr: MultiStatementExpression, stmt_idx: int, stmt: Statement, block: Block | None
    ):
        for idx, stmt_ in enumerate(expr.stmts):
            self._stmt_handler(idx, stmt_, block)
        return super()._handle_MultiStatementExpression(expr_idx, expr, stmt_idx, stmt, block)


class ExpressionUseFinder(AILBlockRewriter):
    """
    Find where each variable is used.

    Additionally, determine if the expression being walked has load expressions inside. Such expressions can only be
    safely folded if there are no Store statements between the expression defining location and its use sites. For
    example, we can only safely fold variable assignments that use Load() when there are no Store()s between the
    assignment and its use site. Otherwise, the loaded expression may get updated later by a Store() statement.

    Here is a real AIL block:

    .. code-block:: none

        v16 = ((int)v23->field_5) + 1 & 255;
        v23->field_5 = ((char)(((int)v23->field_5) + 1 & 255));
        v13 = printf("Recieved packet %d for connection with %d\\n", v16, a0 & 255);

    In this case, folding v16 into the last printf() expression would be incorrect, since v23->field_5 is updated by
    the second statement.
    """

    __slots__ = (
        "has_load",
        "uses",
    )

    def __init__(self):
        super().__init__()
        self.uses: defaultdict[int, set[tuple[Expression, ExpressionLocation | None]]] = defaultdict(set)
        self.has_load = False

    def _handle_expr(
        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Any:
        if isinstance(expr, ailment.Expr.VirtualVariable) and expr.was_reg:
            if not (isinstance(stmt, ailment.Stmt.Assignment) and stmt.dst is expr):
                if block is not None:
                    self.uses[expr.varid].add(
                        (
                            expr,
                            ExpressionLocation(
                                block.addr,
                                block.idx,
                                stmt_idx,
                                expr_idx,
                                phi_stmt=stmt is not None and is_phi_assignment(stmt),
                            ),
                        )
                    )
                else:
                    self.uses[expr.varid].add((expr, None))
            return expr
        return super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Load(self, expr_idx: int, expr: ailment.Expr.Load, stmt_idx: int, stmt: Statement, block: Block | None):
        self.has_load = True
        return super()._handle_Load(expr_idx, expr, stmt_idx, stmt, block)


class ExpressionCounter(SequenceWalker):
    """
    Find all expressions that are assigned once and only used once.
    """

    def __init__(self, node):
        handlers = {
            ConditionalBreakNode: self._handle_ConditionalBreak,
            ConditionNode: self._handle_Condition,
            LoopNode: self._handle_Loop,
            SwitchCaseNode: self._handle_SwitchCase,
            ailment.Block: self._handle_Block,
        }

        # each element in the set is a tuple of (source of the assignment statement, a tuple of unified variables that
        # the current assignment depends on, StatementLocation of the assignment statement, a Boolean variable that
        # indicates if ExpressionUseFinder has succeeded or not)
        self.assignments: defaultdict[Any, set[tuple]] = defaultdict(set)
        self.outerscope_uses: dict[int, set[tuple[Expression, LocationBase | None]]] = {}
        self.all_uses: dict[int, set[tuple[Expression, LocationBase | None]]] = {}
        # inner_scope indicates if we are currently within one of the inner scopes (e.g., a loop). we only collect
        # assignments in the outermost level and stop collecting assignments when we enter inner scopes.
        # we always collect uses, but uses in the outmost scope will be recorded in self.outerscope_uses
        self._outer_scope: bool = True

        super().__init__(handlers)
        self.walk(node)

    def _handle_Statement(self, idx: int, stmt: Statement, node: ailment.Block | LoopNode):
        if not self._outer_scope:
            return
        if isinstance(stmt, ailment.Stmt.Assignment):
            if is_phi_assignment(stmt):
                return
            if isinstance(stmt.dst, ailment.Expr.VirtualVariable) and stmt.dst.was_reg:
                # dependency
                dependency_finder = ExpressionUseFinder()
                dependency_finder.walk_expression(stmt.src)
                dependencies = tuple(dependency_finder.uses)
                self.assignments[stmt.dst.varid].add(
                    (
                        stmt.src,
                        dependencies,
                        StatementLocation(node.addr, node.idx if isinstance(node, ailment.Block) else None, idx),
                        dependency_finder.has_load,
                    )
                )
        if (
            isinstance(stmt, ailment.Stmt.SideEffectStatement)
            and isinstance(stmt.ret_expr, ailment.Expr.VirtualVariable)
            and stmt.ret_expr.was_reg
        ):
            dependency_finder = ExpressionUseFinder()
            dependency_finder.walk_expression(stmt)
            dependencies = tuple(dependency_finder.uses)
            self.assignments[stmt.ret_expr.varid].add(
                (
                    stmt,
                    dependencies,
                    StatementLocation(node.addr, node.idx if isinstance(node, ailment.Block) else None, idx),
                    dependency_finder.has_load,
                )
            )

    def _handle_Block(self, node: ailment.Block, **kwargs):
        # find assignments and uses of variables
        self._collect_uses(node, None)

    def _collect_assignments(self, expr: Expression, node) -> None:
        if not self._outer_scope:
            return
        finder = MultiStatementExpressionAssignmentFinder(self._handle_Statement)
        finder.walk_expression(expr, None, None, node)

    def _collect_uses(self, thing: Expression | Statement | ailment.Block, loc: LocationBase | None):
        use_finder = ExpressionUseFinder()
        if isinstance(thing, ailment.Block):
            for idx, stmt in enumerate(thing.statements):
                self._handle_Statement(idx, stmt, thing)
                use_finder.walk_statement(stmt, block=thing)
        elif isinstance(thing, Statement):
            use_finder.walk_statement(thing)
        else:
            use_finder.walk_expression(thing, stmt_idx=-1)

        for varid, uses in use_finder.uses.items():
            for use in uses:
                # overwrite the location if loc is specified
                content = (use[0], loc) if loc is not None else use

                # update all_uses
                if varid not in self.all_uses:
                    self.all_uses[varid] = set()
                self.all_uses[varid].add(content)

                # update outerscope_uses if we are in the outer scope
                if self._outer_scope:
                    if varid not in self.outerscope_uses:
                        self.outerscope_uses[varid] = set()
                    self.outerscope_uses[varid].add(content)

    def _handle_ConditionalBreak(self, node: ConditionalBreakNode, **kwargs):
        # collect uses on the condition expression
        self._collect_assignments(node.condition, node)
        self._collect_uses(node.condition, ConditionalBreakLocation(node.addr))
        return super()._handle_ConditionalBreak(node, **kwargs)

    def _handle_Condition(self, node: ConditionNode, **kwargs):
        # collect uses on the condition expression
        self._collect_assignments(node.condition, node)
        self._collect_uses(node.condition, ConditionLocation(node.addr))
        return super()._handle_Condition(node, **kwargs)

    def _handle_CascadingCondition(self, node: CascadingConditionNode, **kwargs):
        for idx, (condition, _) in enumerate(node.condition_and_nodes):
            self._collect_assignments(condition, node)
            self._collect_uses(condition, ConditionLocation(node.addr, idx))
        return super()._handle_CascadingCondition(node, **kwargs)

    def _handle_Loop(self, node: LoopNode, **kwargs):
        # collect uses on the condition expression
        if node.initializer is not None:
            self._collect_uses(node.initializer, ConditionLocation(node.addr))
        if node.iterator is not None:
            self._collect_uses(node.iterator, ConditionLocation(node.addr))
        if node.condition is not None:
            self._collect_assignments(node.condition, node)
            self._collect_uses(node.condition, ConditionLocation(node.addr))

        outer_scope = self._outer_scope
        self._outer_scope = False
        super()._handle_Loop(node, **kwargs)
        self._outer_scope = outer_scope

        return None

    def _handle_SwitchCase(self, node: SwitchCaseNode, **kwargs):
        self._collect_uses(node.switch_expr, ConditionLocation(node.addr))
        return super()._handle_SwitchCase(node, **kwargs)


class ExpressionSpotter(VVarUsesCollector):
    """
    ExpressionSpotter collects uses of vvars and existence of Call expressions.
    """

    def __init__(self):
        super().__init__()
        self.has_calls: bool = False
        self.has_loads: bool = False

    def _handle_CallExpr(self, expr_idx: int, expr: Call, stmt_idx: int, stmt: Statement, block: Block | None):
        self.has_calls = True
        return super()._handle_CallExpr(expr_idx, expr, stmt_idx, stmt, block)

    def _handle_Load(self, expr_idx: int, expr: Load, stmt_idx: int, stmt: Statement, block: Block | None):
        self.has_loads = True
        return super()._handle_Load(expr_idx, expr, stmt_idx, stmt, block)


class InterferenceChecker(SequenceWalker):
    """
    Detect for every pair of definition (assignment) - use if there is anything that may interfere with the definition.

    Interferences may be caused by:

    - another call
    - function return
    - store statements
    - load expressions
    - Condition and CascadingCondition nodes
    """

    def __init__(self, assignments: dict[int, Any], uses: dict[int, Any], node):
        handlers = {
            ailment.Block: self._handle_Block,
            ConditionNode: self._handle_Condition,
            ConditionalBreakNode: self._handle_ConditionalBreak,
            SwitchCaseNode: self._handle_SwitchCase,
        }

        super().__init__(handlers, update_seqnode_in_place=False, force_forward_scan=True)
        self._assignments = assignments
        self._uses = uses
        self._assignment_interferences: dict[int, list] = {}
        self.interfered_assignments: set[int] = set()
        self.walk(node)

    def _after_spotting(self, obj, spotter: ExpressionSpotter) -> None:
        if spotter.has_calls or spotter.has_loads:
            # mark all existing assignments as interfered
            for vid in list(self._assignment_interferences):
                self._assignment_interferences[vid].append(obj)
        if set(self._assignment_interferences).intersection(spotter.vvars):
            for used_vvar_id in spotter.vvars:
                if used_vvar_id in self._assignment_interferences:
                    if self._assignment_interferences[used_vvar_id]:
                        self.interfered_assignments.add(used_vvar_id)
                    del self._assignment_interferences[used_vvar_id]

    def _handle_Block(self, node: ailment.Block, **kwargs):
        for stmt in node.statements:
            # deal with uses
            spotter = ExpressionSpotter()
            # special case: we process the call arguments first, then the call itself. this is to allow more expression
            # folding opportunities.
            the_call = None
            if (
                isinstance(stmt, Assignment)
                and isinstance(stmt.src, ailment.Expr.Call)
                and not isinstance(stmt.src.target, str)
            ):
                the_call = stmt.src
            elif isinstance(stmt, ailment.Stmt.SideEffectStatement) and not isinstance(stmt.expr.target, str):
                the_call = stmt.expr
            if the_call is not None:
                assert isinstance(the_call.target, ailment.Stmt.Expression)
                spotter.walk_expression(the_call.target)
                if the_call.args:
                    for arg in the_call.args:
                        spotter.walk_expression(arg)
                self._after_spotting(the_call, spotter)
            spotter.walk_statement(stmt)
            self._after_spotting(stmt, spotter)

            if isinstance(stmt, Return):
                # mark all existing assignments as interfered
                for vid in self._assignment_interferences:
                    self._assignment_interferences[vid].append(stmt)

            if isinstance(stmt, ailment.Stmt.Store):
                # mark all existing assignments as interfered
                for vid in self._assignment_interferences:
                    self._assignment_interferences[vid].append(stmt)

            if isinstance(stmt, ailment.Stmt.SideEffectStatement):
                # mark all existing assignments as interfered
                for vid in self._assignment_interferences:
                    self._assignment_interferences[vid].append(stmt)

            # deal with defs
            if (
                isinstance(stmt, ailment.Stmt.Assignment)
                and isinstance(stmt.dst, ailment.Expr.VirtualVariable)
                and stmt.dst.was_reg
                and stmt.dst.varid in self._assignments
            ):
                # we found this def
                self._assignment_interferences[stmt.dst.varid] = []

            if (
                isinstance(stmt, ailment.Stmt.SideEffectStatement)
                and isinstance(stmt.ret_expr, ailment.Expr.VirtualVariable)
                and stmt.ret_expr.was_reg
                and stmt.ret_expr.variable is not None
                and stmt.ret_expr.varid in self._assignments
            ):
                # we found this def
                self._assignment_interferences[stmt.ret_expr.varid] = []

    def _handle_ConditionalBreak(self, node: ConditionalBreakNode, **kwargs):
        spotter = ExpressionSpotter()
        spotter.walk_expression(node.condition)
        self._after_spotting(node, spotter)
        return super()._handle_ConditionalBreak(node, **kwargs)

    def _handle_Condition(self, node: ConditionNode, **kwargs):
        spotter = ExpressionSpotter()
        spotter.walk_expression(node.condition)
        self._after_spotting(node, spotter)

        # mark all existing assignments as interfered
        for vid in self._assignment_interferences:
            self._assignment_interferences[vid].append(node)

        return super()._handle_Condition(node, **kwargs)

    def _handle_CascadingCondition(self, node: CascadingConditionNode, **kwargs):
        spotter = ExpressionSpotter()
        for cond, _ in node.condition_and_nodes:  # pylint:disable=consider-using-enumerate
            spotter.walk_expression(cond)
            self._after_spotting(node, spotter)

        # mark all existing assignments as interfered
        for vid in self._assignment_interferences:
            self._assignment_interferences[vid].append(node)

        return super()._handle_CascadingCondition(node, **kwargs)

    def _handle_Loop(self, node: LoopNode, **kwargs):
        spotter = ExpressionSpotter()

        # iterator
        if node.iterator is not None:
            spotter.walk_statement(node.iterator)

        # initializer
        if node.initializer is not None:
            spotter.walk_statement(node.initializer)

        # condition
        if node.condition is not None:
            spotter.walk_expression(node.condition)

        self._after_spotting(node, spotter)

        return super()._handle_Loop(node, **kwargs)

    def _handle_SwitchCase(self, node: SwitchCaseNode, **kwargs):
        spotter = ExpressionSpotter()
        spotter.walk_expression(node.switch_expr)
        self._after_spotting(node, spotter)
        return super()._handle_SwitchCase(node, **kwargs)


class ExpressionReplacer(AILBlockRewriter):
    def __init__(self, assignments: dict[int, Any], uses: dict[int, Any]):
        super().__init__()
        self._assignments = assignments
        self._uses = uses

    def _handle_MultiStatementExpression(  # type: ignore
        self, expr_idx, expr: MultiStatementExpression, stmt_idx: int, stmt: Statement, block: Block | None
    ) -> Expression | None:
        changed = False
        new_statements = []
        for idx, stmt_ in enumerate(expr.stmts):
            if (
                isinstance(stmt_, Assignment)
                and isinstance(stmt_.dst, ailment.Expr.VirtualVariable)
                and stmt_.dst.was_reg
                and stmt_.dst.variable is not None
            ) and stmt_.dst.variable in self._assignments:
                # remove this statement
                changed = True
                continue

            new_stmt = self._handle_stmt(idx, stmt_, None)
            if new_stmt is not None and new_stmt is not stmt_:
                changed = True
                if isinstance(new_stmt, Assignment) and new_stmt.src.likes(new_stmt.dst):
                    # this statement is simplified into reg = reg. ignore it
                    continue
                new_statements.append(new_stmt)
            else:
                new_statements.append(stmt_)

        new_expr = self._handle_expr(0, expr.expr, stmt_idx, stmt, block)
        if new_expr is not None and new_expr is not expr.expr:
            changed = True
        else:
            new_expr = expr.expr

        if changed:
            if not new_statements:
                # it is no longer a multi-statement expression
                return new_expr  # type: ignore
            expr_ = expr.copy()
            expr_.expr = new_expr
            expr_.stmts = new_statements
            return expr_
        return expr

    def _handle_Assignment(self, stmt_idx: int, stmt: Assignment, block: Block | None):
        # override the base handler and make sure we do not replace .dst with a Call expression or an ITE expression

        if is_phi_assignment(stmt):
            return stmt

        if isinstance(stmt.dst, VirtualVariable) and stmt.dst.varid in self._assignments:
            return stmt

        changed = False

        dst = self._handle_expr(0, stmt.dst, stmt_idx, stmt, block)
        if dst is not stmt.dst and not isinstance(dst, (Call, ITE)):
            changed = True
        else:
            dst = stmt.dst
        assert isinstance(dst, Atom)

        src = self._handle_expr(1, stmt.src, stmt_idx, stmt, block)
        if src is not stmt.src:
            changed = True
        else:
            src = stmt.src

        if changed:
            return Assignment(stmt.idx, dst, src, **stmt.tags)
        return stmt

    def _handle_expr(
        self, expr_idx: int, expr: Expression, stmt_idx: int, stmt: Statement | None, block: Block | None
    ) -> Expression:
        if isinstance(expr, ailment.Expr.VirtualVariable) and expr.was_reg and expr.varid in self._uses:
            replace_with, _ = self._assignments[expr.varid]
            return replace_with
        return super()._handle_expr(expr_idx, expr, stmt_idx, stmt, block)


class ExpressionFolder(SequenceWalker):
    def __init__(self, assignments: dict[int, Any], uses: dict[int, Any], node):
        handlers = {
            ailment.Block: self._handle_Block,
            ConditionNode: self._handle_Condition,
            ConditionalBreakNode: self._handle_ConditionalBreak,
            SwitchCaseNode: self._handle_SwitchCase,
            LoopNode: self._handle_Loop,
        }

        super().__init__(handlers)
        self._assignments = assignments
        self._uses = uses
        self.walk(node)

    def _handle_Block(self, node: ailment.Block, **kwargs):
        # Walk the block to remove each assignment and replace uses of each variable
        new_stmts = []
        for stmt in node.statements:
            if (
                isinstance(stmt, ailment.Stmt.Assignment)
                and isinstance(stmt.dst, ailment.Expr.VirtualVariable)
                and stmt.dst.was_reg
                and stmt.dst.varid in self._assignments
            ):
                # remove this statement
                continue
            if (
                isinstance(stmt, ailment.Stmt.SideEffectStatement)
                and isinstance(stmt.ret_expr, ailment.Expr.VirtualVariable)
                and stmt.ret_expr.was_reg
                and stmt.ret_expr.variable is not None
                and stmt.ret_expr.varid in self._assignments
            ):
                # remove this statement
                continue
            new_stmts.append(stmt)
        node.statements = new_stmts

        # Walk the block to replace the use of each variable
        replacer = ExpressionReplacer(self._assignments, self._uses)
        replacer.walk(node)

    def _handle_ConditionalBreak(self, node: ConditionalBreakNode, **kwargs):
        replacer = ExpressionReplacer(self._assignments, self._uses)
        r = replacer.walk_expression(node.condition)
        if r is not None and r is not node.condition:
            node.condition = r
        return super()._handle_ConditionalBreak(node, **kwargs)

    def _handle_Condition(self, node: ConditionNode, **kwargs):
        replacer = ExpressionReplacer(self._assignments, self._uses)
        r = replacer.walk_expression(node.condition)
        if r is not None and r is not node.condition:
            node.condition = r
        return super()._handle_Condition(node, **kwargs)

    def _handle_CascadingCondition(self, node: CascadingConditionNode, **kwargs):
        replacer = ExpressionReplacer(self._assignments, self._uses)
        for idx in range(len(node.condition_and_nodes)):  # pylint:disable=consider-using-enumerate
            cond, _ = node.condition_and_nodes[idx]
            r = replacer.walk_expression(cond)
            if r is not None and r is not cond:
                node.condition_and_nodes[idx] = (r, node.condition_and_nodes[idx][1])
        return super()._handle_CascadingCondition(node, **kwargs)

    def _handle_Loop(self, node: LoopNode, **kwargs):
        replacer = ExpressionReplacer(self._assignments, self._uses)

        # iterator
        if node.iterator is not None:
            r = replacer.walk_statement(node.iterator)
            if r is not None and r is not node.iterator:
                node.iterator = r

        # initializer
        if node.initializer is not None:
            r = replacer.walk_statement(node.initializer)
            if r is not None and r is not node.initializer:
                node.initializer = r

        # condition
        if node.condition is not None:
            r = replacer.walk_expression(node.condition)
            if r is not None and r is not node.condition:
                node.condition = r

        # again, do not replace into the loop body
        return None

    def _handle_SwitchCase(self, node: SwitchCaseNode, **kwargs):
        replacer = ExpressionReplacer(self._assignments, self._uses)

        r = replacer.walk_expression(node.switch_expr)
        if r is not None and r is not node.switch_expr:
            node.switch_expr = r

        return super()._handle_SwitchCase(node, **kwargs)


class StoreStatementFinder(SequenceWalker):
    """
    Determine if there are any Store statements between two given statements.

    This class overrides _handle_Sequence() and _handle_MultiNode() to ensure they traverse nodes from top to bottom.
    """

    def __init__(self, node, intervals: Iterable[tuple[StatementLocation, LocationBase]]):
        handlers = {
            ConditionNode: self._handle_Condition,
            CascadingConditionNode: self._handle_CascadingCondition,
            ConditionalBreakNode: self._handle_ConditionalBreak,
            ailment.Block: self._handle_Block,
        }

        self._intervals = intervals

        self._start_to_ends: defaultdict[StatementLocation, set[LocationBase]] = defaultdict(set)
        self._end_to_starts: defaultdict[LocationBase, set[StatementLocation]] = defaultdict(set)
        self.interval_to_hasstore: dict[tuple[StatementLocation, StatementLocation], bool] = {}
        for start, end in intervals:
            self._start_to_ends[start].add(end)
            self._end_to_starts[end].add(start)

        self._active_intervals = set()

        super().__init__(handlers)
        self.walk(node)

    def _handle_Sequence(self, node, **kwargs):
        i = 0
        while i < len(node.nodes):
            node_ = node.nodes[i]
            self._handle(node_, parent=node, index=i)
            i += 1

    def _handle_MultiNode(self, node, **kwargs):
        i = 0
        while i < len(node.nodes):
            node_ = node.nodes[i]
            self._handle(node_, parent=node, index=i)
            i += 1

    def _handle_Block(self, node: ailment.Block, **kwargs):
        stmt_loc = StatementLocation(node.addr, node.idx, None)
        for idx, stmt in enumerate(node.statements):
            stmt_loc.stmt_idx = idx
            if stmt_loc in self._start_to_ends:
                for end in self._start_to_ends[stmt_loc]:
                    self._active_intervals.add((stmt_loc.copy(), end))
            if stmt_loc in self._end_to_starts:
                for start in self._end_to_starts[stmt_loc]:
                    self._active_intervals.discard((start, stmt_loc))
            if isinstance(stmt, ailment.Stmt.Store):
                for interval in self._active_intervals:
                    self.interval_to_hasstore[interval] = True

    def _handle_Condition(self, node, **kwargs):
        cond_loc = ConditionLocation(node.addr)
        if cond_loc in self._end_to_starts:
            for start in self._end_to_starts[cond_loc]:
                self._active_intervals.discard((start, cond_loc))
        super()._handle_Condition(node, **kwargs)

    def _handle_CascadingCondition(self, node: CascadingConditionNode, **kwargs):
        cond_loc = ConditionLocation(node.addr, None)
        for idx in range(len(node.condition_and_nodes)):
            cond_loc.case_idx = idx
            if cond_loc in self._end_to_starts[cond_loc]:
                for start in self._end_to_starts[cond_loc]:
                    self._active_intervals.discard((start, cond_loc))
        super()._handle_CascadingCondition(node, **kwargs)

    def _handle_ConditionalBreak(self, node: ConditionalBreakNode, **kwargs):
        cond_break_loc = ConditionalBreakLocation(node.addr)
        if cond_break_loc in self._end_to_starts:
            for start in self._end_to_starts[cond_break_loc]:
                self._active_intervals.discard((start, cond_break_loc))
        super()._handle_ConditionalBreak(node, **kwargs)

    def has_store(self, start: StatementLocation, end: StatementLocation) -> bool:
        return self.interval_to_hasstore.get((start, end), False)



# pylint:disable=too-many-boolean-expressions
from __future__ import annotations

from angr.ailment.expression import BinaryOp, Const, Convert, Expression

from .base import PeepholeOptimizationExprBase


class ConcatSimplifier(PeepholeOptimizationExprBase):
    """
    Simplify Concat expressions:
    - (a CONCAT b) >> bits(b)  =>  a  (high-part extraction)
    - (a CONCAT b) & mask  =>  b  (low-part extraction, where mask = (1 << bits(b)) - 1)
    - (a >> (bits-1)) CONCAT a  =>  Convert(a, signed, 2*bits)  (sign-extension)
    - 0 CONCAT a  =>  Convert(a, unsigned, 2*bits)  (zero-extension)
    - Convert(a CONCAT b, to_bits=bits(b))  =>  b  (truncate to low part)
    """

    __slots__ = ()

    NAME = "Simplify Concat expressions"
    expr_classes = (BinaryOp, Convert)

    def optimize(self, expr: BinaryOp | Convert, stmt_idx: int | None = None, block=None, **kwargs):
        if isinstance(expr, BinaryOp):
            if expr.op == "Concat":
                return self._optimize_concat(expr, stmt_idx, block)
            if expr.op in {"Shr", "Sar"}:
                return self._optimize_shr_concat(expr)
            if expr.op == "And":
                return self._optimize_and_concat(expr)
        elif isinstance(expr, Convert):
            return self._optimize_convert_concat(expr)
        return None

    def _optimize_concat(self, expr: BinaryOp, stmt_idx: int | None = None, block=None) -> Expression | None:
        """
        Simplify Concat expressions that represent sign/zero extension.
        """
        high, low = expr.operands

        # Pattern: 0 CONCAT a  =>  Convert(a, unsigned, 2*bits)
        if isinstance(high, Const) and high.value == 0:
            return Convert(
                expr.idx,
                low.bits,
                expr.bits,
                False,  # unsigned
                low,
                **expr.tags,
            )

        # Pattern: (a >> (bits-1)) CONCAT a  =>  Convert(a, signed, 2*bits)
        if (
            isinstance(high, BinaryOp)
            and high.op == "Sar"
            and isinstance(high.operands[1], Const)
            and high.operands[1].value == low.bits - 1
            and high.operands[0].likes(low)
        ):
            return Convert(
                expr.idx,
                low.bits,
                expr.bits,
                True,  # signed
                low,
                **expr.tags,
            )

        # Pattern: ((signed)Conv(a) >> (bits-1)) CONCAT a  =>  Convert(a, signed, 2*bits)
        # This handles cases where the sign bit extraction uses a converted value
        if (
            isinstance(high, BinaryOp)
            and high.op == "Sar"
            and isinstance(high.operands[1], Const)
            and isinstance(high.operands[0], Convert)
        ):
            conv = high.operands[0]
            shift_amt = high.operands[1].value
            if conv.is_signed and conv.operand.likes(low) and shift_amt == conv.to_bits - 1:
                return Convert(
                    expr.idx,
                    low.bits,
                    expr.bits,
                    True,  # signed
                    low,
                    **expr.tags,
                )

        # Pattern: v CONCAT a where v = a >> (bits-1)  =>  Convert(a, signed, 2*bits)
        # This handles cases where the sign extension is split across statements
        if stmt_idx is not None and block is not None:
            high_def = self.find_definition(high, stmt_idx, block)
            if (
                isinstance(high_def, BinaryOp)
                and high_def.op == "Sar"
                and isinstance(high_def.operands[1], Const)
                and high_def.operands[1].value == low.bits - 1
                and high_def.operands[0].likes(low)
            ):
                return Convert(
                    expr.idx,
                    low.bits,
                    expr.bits,
                    True,  # signed
                    low,
                    **expr.tags,
                )

        return None

    @staticmethod
    def _optimize_shr_concat(expr: BinaryOp) -> Expression | None:
        """
        Simplify (a CONCAT b) >> bits(b)  =>  a  (high-part extraction)
        """
        if not isinstance(expr.operands[1], Const):
            return None

        shift_amt = expr.operands[1].value
        inner = expr.operands[0]

        # Handle Convert wrapping the Concat
        outer_convert = None
        if isinstance(inner, Convert):
            outer_convert = inner
            inner = inner.operand

        if not isinstance(inner, BinaryOp) or inner.op != "Concat":
            return None

        high, low = inner.operands

        # Check if we're extracting exactly the high part
        if shift_amt != low.bits:
            return None
        result = high

        # Apply conversion if needed
        if expr.bits != result.bits:
            result = Convert(
                expr.idx,
                result.bits,
                expr.bits,
                outer_convert.is_signed if outer_convert is not None else False,
                result,
                **expr.tags,
            )

        return result

    @staticmethod
    def _optimize_and_concat(expr: BinaryOp) -> Expression | None:
        """
        Simplify (a CONCAT b) & mask  =>  b  (low-part extraction)
        where mask = (1 << bits(b)) - 1
        """
        if not isinstance(expr.operands[1], Const):
            return None

        mask = expr.operands[1].value
        inner = expr.operands[0]

        # Handle Convert wrapping the Concat
        outer_convert = None
        if isinstance(inner, Convert):
            outer_convert = inner
            inner = inner.operand

        if not isinstance(inner, BinaryOp) or inner.op != "Concat":
            return None

        _high, low = inner.operands

        # Check if the mask extracts exactly the low part
        expected_mask = (1 << low.bits) - 1
        if mask != expected_mask:
            return None

        # The result should be the low part
        result = low

        if outer_convert is not None:
            if outer_convert.to_bits != low.bits:
                result = Convert(
                    outer_convert.idx,
                    low.bits,
                    outer_convert.to_bits,
                    False,  # zero-extend since we masked
                    result,
                    **outer_convert.tags,
                )
        elif expr.bits != low.bits:
            result = Convert(
                expr.idx,
                low.bits,
                expr.bits,
                False,  # zero-extend since we masked
                result,
                **expr.tags,
            )

        return result

    @staticmethod
    def _optimize_convert_concat(expr: Convert) -> Expression | None:
        """
        Simplify Convert(a CONCAT b, to_bits=bits(b))  =>  b  (truncate to low part)
        """
        if expr.from_bits <= expr.to_bits:
            return None

        inner = expr.operand
        if not isinstance(inner, BinaryOp) or inner.op != "Concat":
            return None

        _high, low = inner.operands

        # Check if we're truncating to exactly the low part size
        if expr.to_bits == low.bits:
            return low

        # Check if we're truncating to less than the low part
        if expr.to_bits < low.bits:
            return Convert(
                expr.idx,
                low.bits,
                expr.to_bits,
                expr.is_signed,
                low,
                **expr.tags,
            )

        return None



from __future__ import annotations
from typing import Any, TYPE_CHECKING

from .clinic import Clinic
from .structured_codegen import BaseStructuredCodeGenerator

if TYPE_CHECKING:
    from angr.analyses.decompiler.optimization_passes.expr_op_swapper import OpDescriptor
    from angr.analyses.typehoon.typevars import TypeVariable, TypeConstraint


class DecompilationCache:
    """
    Caches key data structures that can be used later for refining decompilation results, such as retyping variables.
    """

    __slots__ = (
        "addr",
        "arg_vvars",
        "binop_operators",
        "clinic",
        "codegen",
        "errors",
        "func_typevar",
        "function_summary",
        "ite_exprs",
        "notes",
        "parameters",
        "stack_offset_typevars",
        "stackvar_max_sizes",
        "type_constraints",
        "var_to_typevar",
    )

    def __init__(self, addr):
        self.parameters: dict[str, Any] = {}
        self.addr = addr
        self.type_constraints: dict[TypeVariable, set[TypeConstraint]] | None = None
        self.arg_vvars: dict | None = None
        self.func_typevar: TypeVariable | None = None
        self.var_to_typevar: dict | None = None
        self.stackvar_max_sizes: dict | None = None
        self.stack_offset_typevars: dict | None = None
        self.codegen: BaseStructuredCodeGenerator | None = None
        self.clinic: Clinic | None = None
        self.ite_exprs: set[tuple[int, Any]] | None = None
        self.binop_operators: dict[OpDescriptor, str] | None = None
        self.errors: list[str] = []
        self.function_summary: str | None = None
        self.notes: dict[str, str] = {}

    @property
    def local_types(self):
        if self.clinic is None or self.clinic.variable_kb is None:
            return None
        return self.clinic.variable_kb.variables[self.addr].types



from __future__ import annotations

from angr.ailment import Block
from .sequence_walker import SequenceWalker
from .structuring.structurer_nodes import BaseNode, SequenceNode, MultiNode


class NodeReplacer(SequenceWalker):
    """
    Replaces nodes in a node with new nodes based on a mapping.
    """

    def __init__(self, root: BaseNode, replacements: dict) -> None:
        super().__init__(update_seqnode_in_place=False)

        self.root = root
        self.replacements = replacements
        self.result: BaseNode = self.walk(self.root)  # type: ignore

    def _handle(self, node: BaseNode, **kwargs):
        return self.replacements[node] if node in self.replacements else super()._handle(node, **kwargs)

    def _handle_MultiNode(self, node: MultiNode, **kwargs):
        changed = False
        nodes_copy = list(node.nodes)

        i = len(nodes_copy) - 1
        has_non_block = False
        while i > -1:
            node_ = nodes_copy[i]
            new_node = self._handle(node_, parent=node, index=i)
            if new_node is not None:
                changed = True
                nodes_copy[i] = new_node
                if not isinstance(new_node, Block):
                    has_non_block = True
            i -= 1
        if not changed:
            return None
        if has_non_block:
            return SequenceNode(node.addr, nodes=nodes_copy)
        return MultiNode(nodes_copy, addr=node.addr, idx=node.idx)



from __future__ import annotations
import angr


class exit(angr.SimProcedure):  # pylint:disable=redefined-builtin
    # pylint:disable=arguments-differ

    NO_RET = True

    def run(self, exit_code):
        self.exit(exit_code)



from __future__ import annotations
from .base import SimSootValue


class SimSootValue_StringRef(SimSootValue):
    __slots__ = ["id", "type"]

    def __init__(self, heap_alloc_id):
        self.id = f"{heap_alloc_id}.string"
        self.type = "java.lang.String"

    def __repr__(self):
        return self.id

    def __eq__(self, other):
        return isinstance(other, SimSootValue_StringRef) and self.id == other.id and self.type == other.type

    @classmethod
    def from_sootvalue(cls, soot_value, state):
        raise NotImplementedError

    @staticmethod
    def new_string(state, value):
        """
        Allocate and initialize a new string in the context of the state passed.

        The method returns the reference to the newly allocated string

        :param state: angr state where we want to allocate the string
        :type SimState
        :param value: value of the string to initialize
        :type claripy.String

        :return SimSootValue_StringRef
        """
        str_ref = SimSootValue_StringRef(state.memory.get_new_uuid())
        state.memory.store(str_ref, value)
        return str_ref



from __future__ import annotations
import claripy

import angr


class UserHook(angr.SimProcedure):
    NO_RET = True

    # pylint: disable=arguments-differ
    def run(self, user_func=None, length=None):
        result = user_func(self.state)
        if result is None:
            jumpkind = "Ijk_NoHook" if length == 0 else "Ijk_Boring"
            self.successors.add_successor(self.state, self.state.addr + length, claripy.true(), jumpkind)
        else:
            for state in result:
                self.successors.add_successor(state, state.addr, state.scratch.guard, state.history.jumpkind)



from __future__ import annotations
import claripy

import angr


class isgraph(angr.SimProcedure):
    # pylint: disable=arguments-differ, missing-class-docstring
    def run(self, c):
        is_graph = claripy.And(c >= 33, c <= 126)
        int_size = self.arch.sizeof["int"]  # type: ignore[reportOptionalMemberAccess]

        return claripy.If(is_graph, claripy.BVV(1, int_size), claripy.BVV(0, int_size))
